# Result of a record's JSON export
DETAIL = {
    "dty_ID": 1113,
    "value": {
        "file": {
            "ulf_ID": "1",
            "fullPath": None,
            "ulf_ExternalFileReference": "https://upload.wikimedia.org/wikipedia/commons/9/9a/Gull_portrait_ca_usa.jpg",
            "fxm_MimeType": "image/jpeg",
            "ulf_PreferredSource": "local",
            "ulf_OrigFileName": "_remote",
            "ulf_FileSizeKB": "0",
            "ulf_ObfuscatedFileID": "286ccc9935b4ac7d2ba372b561c5ce9a5ae609dd",
            "ulf_Description": "",
            "ulf_Added": "2024-04-05 19:48:05",
            "ulf_MimeExt": "jpe",
            "ulf_Caption": "seagull from wikimedia images",
            "ulf_Copyright": "",
            "ulf_Copyowner": "",
            "ulf_Parameters": None,
            "ulf_WhoCanView": None,
        },
        "fileid": "286ccc9935b4ac7d2ba372b561c5ce9a5ae609dd",
    },
    "fieldName": "file or media url",
    "fieldType": "file",
    "conceptID": "",
}



"""Heurist API session"""

import requests
from heurist.api.client import HeuristAPIClient
from heurist.api.constants import READTIMEOUT
from heurist.api.exceptions import AuthenticationError
from requests import Session


class HeuristAPIConnection:
    def __init__(
        self,
        db: str,
        login: str,
        password: str,
        read_timeout: int = READTIMEOUT,
        post_timeout: int = 10,
    ) -> None:
        """
        Session context for a connection to the Heurist server.

        Args:
            db (str): Heurist database name.
            login (str): Username.
            password (str): User's password.
            read_timeout (int): Seconds to wait before raising a ReadTimeout.
            post_timeout (int): Seconds to wait before raising an error when \
                establishing a login connection.

        Raises:
            e: If the requests method fails, raise that exception.
            AuthenticationError: If the Heurist server returns a bad status code, \
                raise an exception.
        """

        self.db = db
        self.__login = login
        self.__password = password
        self._readtimeout = read_timeout
        self._posttimeout = post_timeout

    def __enter__(self) -> Session:
        self.session = requests.Session()
        url = "https://heurist.huma-num.fr/heurist/api/login"

        body = {
            "db": self.db,
            "login": self.__login,
            "password": self.__password,
        }
        try:
            response = self.session.post(url=url, data=body, timeout=self._posttimeout)
        except requests.exceptions.ConnectTimeout as e:
            print(
                "\nUnable to log in to Heurist Huma-Num server. \
                  Connection timed out."
            )
            raise e
        if response.status_code != 200:
            message = response.json()["message"]
            e = AuthenticationError(message)
            raise SystemExit(e)

        return HeuristAPIClient(
            database_name=self.db,
            session=self.session,
            timeout_seconds=self._readtimeout,
        )

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.session.close()



"""Heurist API client"""

import json
from typing import ByteString, Literal

import requests
from heurist.api.constants import MAX_RETRY, READTIMEOUT
from heurist.api.exceptions import APIException, ReadTimeout
from heurist.api.url_builder import URLBuilder
from heurist.api.utils import log_attempt_number
from tenacity import (
    RetryError,
    retry,
    retry_if_exception_type,
    stop_after_attempt,
)


class HeuristAPIClient:
    """
    Client for Heurist API.
    """

    def __init__(
        self,
        database_name: str,
        session: requests.Session,
        timeout_seconds: int | None = READTIMEOUT,
    ) -> None:
        self.database_name = database_name
        self.url_builder = URLBuilder(database_name=database_name)
        self.session = session
        self.timeout = timeout_seconds

    @retry(
        retry=retry_if_exception_type(requests.exceptions.ReadTimeout),
        stop=stop_after_attempt(MAX_RETRY),
        after=log_attempt_number,
    )
    def call_heurist_api(self, url: str) -> ByteString | None:
        response = self.session.get(url, timeout=(self.timeout))
        return response

    def get_response_content(self, url: str) -> ByteString | None:
        """Request resources from the Heurist server.

        Args:
            url (str): Heurist API entry point.

        Returns:
            ByteString | None: Binary response returned from Heurist server.
        """

        try:
            response = self.call_heurist_api(url=url)
        except RetryError:
            e = ReadTimeout(url=url, timeout=self.timeout)
            raise SystemExit(e)
        if not response:
            e = APIException("No response.")
            raise SystemExit(e)
        elif response.status_code != 200:
            e = APIException(f"Status {response.status_code}")
            raise SystemExit(e)
        elif "Cannot connect to database" == response.content.decode("utf-8"):
            e = APIException("Could not connect to database.")
            raise SystemExit(e)
        else:
            return response.content

    def get_records(
        self,
        record_type_id: int,
        form: Literal["xml", "json"] = "json",
        users: tuple[int] = (),
    ) -> bytes | list | None:
        """Request all records of a certain type and in a certain data format.

        Args:
            record_type_id (int): Heurist ID of targeted record type.
            form (Literal["xml", "json"], optional): Data format for requested
                records. Defaults to "json".
            users (tuple): Array of IDs of users who added the target records.

        Returns:
            bytes | list | None: If XML, binary response returned from Heurist
                server, else JSON array.
        """

        url = self.url_builder.get_records(
            record_type_id=record_type_id, form=form, users=users
        )
        if form == "json":
            content = self.get_response_content(url)
            json_string = content.decode("utf-8")
            all_records = json.loads(json_string)["heurist"]["records"]
            # Filter out linked records of a not the target type
            correct_ids = [
                r for r in all_records if r["rec_RecTypeID"] == str(record_type_id)
            ]
            # Filter out records by non-targeted users
            if users and len(users) > 0:
                return [r for r in correct_ids if int(r["rec_AddedByUGrpID"]) in users]
            else:
                return correct_ids
        else:
            return self.get_response_content(url)

    def get_structure(self) -> bytes | None:
        """Request the Heurist database's overall structure in XML format.

        Returns:
            bytes | list | None: If XML, binary response returned from Heurist server,
            else JSON array.
        """
        url = self.url_builder.get_db_structure()
        return self.get_response_content(url)

    def get_relationship_markers(
        self, form: Literal["xml", "json"] = "xml"
    ) -> bytes | list | None:
        return self.get_records(record_type_id=1, form=form)



import logging
import os

from heurist.log.constants import VALIDATION_LOG
from heurist.models.dynamic.annotation import PydanticField
from heurist.models.dynamic.type import FieldType
from heurist.validators.detail_validator import DetailValidator
from heurist.validators.exceptions import RepeatedValueInSingularDetailType
from pydantic import BaseModel

handlers = [logging.FileHandler(filename=VALIDATION_LOG, mode="w")]
if os.getenv("HEURIST_STREAM_LOG") == "True":
    handlers.append(logging.StreamHandler())

logging.basicConfig(
    encoding="utf-8",
    format="{asctime} - {levelname} - {message}",
    style="{",
    datefmt="%Y-%m-%d %H:%M",
    handlers=handlers,
)


def list_plural_fields(pydantic_model: BaseModel) -> list:
    return [
        v.description
        for v in pydantic_model.model_fields.values()
        if repr(v.annotation).startswith("list")
    ]


class RecordValidator:
    def __init__(
        self, pydantic_model: BaseModel, records: list[dict], rty_ID: int
    ) -> None:
        self.pydantic_model = pydantic_model
        self._rty_ID = rty_ID
        self._records = records
        self._index = 0
        self._plural_fields = list_plural_fields(pydantic_model=self.pydantic_model)

    def is_plural(self, dty_ID: int) -> bool:
        if dty_ID in self._plural_fields:
            return True

    def __iter__(self):
        return self

    def __next__(self) -> BaseModel:
        if self._index < len(self._records):
            record = self._records[self._index]
            self._index += 1
            # If the record isn't of the record type for this model, skip it.
            if record["rec_RecTypeID"] != self._rty_ID:
                pass
            # Otherwise, process the record's details into key-value pairs that
            # will be loaded into the Pydantic model.
            kwargs = self.flatten_details_to_dynamic_pydantic_fields(record)
            # Return a validated Pydantic model.
            return self.pydantic_model.model_validate(kwargs)
        else:
            raise StopIteration

    @classmethod
    def aggregate_details_by_type(cls, details: list[dict]) -> dict:
        # Set up an index for all the types of details in this record's
        # sequence of details.
        index = {d["dty_ID"]: [] for d in details}
        # According to its type, add each detail to its respective list in the index.
        [index[d["dty_ID"]].append(d) for d in details]
        # Return the index of aggregated details.
        return index

    def flatten_details_to_dynamic_pydantic_fields(self, record: dict) -> dict:
        detail_type_index = self.aggregate_details_by_type(record["details"])
        # To the list of key-value pairs, add the record's H-ID and its type ID
        record_id = record["rec_ID"]
        kwargs = {
            "rec_ID": record_id,
            "rec_RecTypeID": record["rec_RecTypeID"],
        }
        for dty_ID, details in detail_type_index.items():
            # Determine if this detail type is allowed to have multiple values.
            repeats = self.is_plural(dty_ID=dty_ID)

            # If this detail is not supposed to be repeateable but Heurist allowed more
            # than 1 value to be saved in the field, raise an error.
            if not repeats and len(details) > 1:
                warning = RepeatedValueInSingularDetailType(
                    type_id=record["rec_RecTypeID"],
                    record_id=record_id,
                    field_name=details[0]["fieldName"],
                    value_count=len(details),
                )
                logging.warning(warning)
                continue

            # Get the validation alias for this kwarg's key
            key = PydanticField._get_validation_alias(dty_ID=dty_ID)

            # Convert the detail's metadata to a flat value.
            values = []
            for detail in details:
                v = DetailValidator.convert(detail=detail)
                values.append(v)

            # Check the number of validated metadata against what is permissible for
            # this detail type according to the Heurist schema.
            value = self.validate_for_repeatable_values(repeats=repeats, values=values)

            # If the validation failed, do not add this detail type to the set of
            # kwargs for the Pydantic model. Let the model's default value be used
            # for this missing / invalid metadata.
            if not value:
                continue

            # Add this detail type's alias and validated value(s) to the set of kwargs.
            kwargs.update({key: value})

            # If the detail is a Term, add an additional field for the foreign key.
            if FieldType.from_detail(details[0]) == "enum":
                # To this detail type's validation alias, which is associated with the
                # term's label, append a suffix to distinguish it as a supplemental
                # field to hold the foreign key.
                key += PydanticField.trm_validation_alias_suffix
                # Into a list, extract each detail's foreign key, which is in "value."
                values = []
                for detail in details:
                    values.append(detail["value"])

                value = self.validate_for_repeatable_values(
                    repeats=repeats, values=values
                )

                # The previous if-condition should have already confirmed that this
                # group of deatils are valid. Therefore, they can be added directly
                # to the kwargs.
                kwargs.update({key: value})

        # Return the flat key-value pairs for the Pydantic model's fields.
        return kwargs

    @classmethod
    def validate_for_repeatable_values(
        cls, repeats: bool, values: list
    ) -> list | dict | None:
        # If the detail type is not repeatable, extract the first dictionary.
        if not repeats and len(values) > 0:
            return values[0]
        # If the detail type is repeatable, send the list of values, which can
        # be an empty list--as this should be the default value for this field
        # annotation.
        elif repeats:
            return values



import unittest
from pathlib import Path

import duckdb
from heurist.api.credentials import CredentialHandler
from heurist.cli.load import load_command


class DownloadCommand(unittest.TestCase):
    def setUp(self):
        try:
            self.credentials = CredentialHandler(database_name="api_dev")
        except SystemExit:
            self.skipTest(
                "Connection could not be established.\nCannot test client without \
                    database connection."
            )
        self.database_connection_path = Path(__file__).parent.joinpath("test.db")

    def tearDown(self):
        self.database_connection_path.unlink()

    def test_database_serialization(self):
        load_command(
            credentials=self.credentials,
            duckdb_database_connection_path=self.database_connection_path,
        )

        conn = duckdb.connect(str(self.database_connection_path))
        expected = 5  # 5 non-record tables are created
        actual = len(conn.execute("show tables;").fetchall())
        self.assertGreater(actual, expected)


if __name__ == "__main__":
    unittest.main()



from typing import Any, Optional

from heurist.models.dynamic.type import FieldType
from heurist.sql.sql_safety import SafeSQLName
from pydantic import Field


class PydanticField:
    trm_validation_alias_suffix = "_TRM"

    def __init__(
        self, dty_ID: int, rst_DisplayName: str, dty_Type: str, rst_MaxValues: int
    ):
        """
        Using information of 1 detail (data field) from a Heurist record, build a \
            Pydantic data field annotation.

        Args:
            dty_ID (int): Detail's ID.
            rst_DisplayName (str): Name of the detail displayed in Heurist.
            dty_Type (str): Detail's data type.
            rst_MaxValues (int): Heurist indicator if the detail can be repeated.
        """
        self.dty_ID = dty_ID
        self.rst_DisplayName = rst_DisplayName
        self.dty_Type = dty_Type
        self.rst_MaxValues = rst_MaxValues

    @classmethod
    def _get_validation_alias(cls, dty_ID: int) -> str:
        return f"DTY{dty_ID}"

    @property
    def validation_alias(self) -> str:
        return self._get_validation_alias(dty_ID=self.dty_ID)

    @property
    def serlialization_alias(self) -> str:
        return SafeSQLName().create_column_name(
            field_name=self.rst_DisplayName, field_type=self.dty_Type
        )

    @property
    def pydantic_type(self) -> Any:
        fieldtype = FieldType.to_pydantic(datatype=self.dty_Type)
        if self._is_type_repeatable():
            return list[fieldtype]
        else:
            return fieldtype

    def build_field(self) -> dict:
        """
        Build a Pydantic field annotation for a detail whose value will simply be the \
            result of the `RecordDetailConverter`, meaning not a date and not a \
            vocabulary term.

        Returns:
            dict: Pydantic field annotation.
        """

        return self._compose_annotation(
            validation_alias=self.validation_alias,
            serialization_alias=self.serlialization_alias,
            pydantic_type=self.pydantic_type,
        )

    def build_term_fk(self) -> dict:
        """
        Build a Pydantic field annotation for a foreign key reference to the vocabulary\
            term in the constructed database's trm table. This field is written to the \
            Pydantic model in addition to a column for the term that simply has the \
            label.

        Returns:
            dict: Pydantic field annotation.
        """

        validation_alias = self.validation_alias + self.trm_validation_alias_suffix
        serialization_alias = self.serlialization_alias + " TRM-ID"
        if self._is_type_repeatable():
            pydantic_type = list[Optional[int]]
        else:
            pydantic_type = Optional[int]

        return self._compose_annotation(
            validation_alias=validation_alias,
            serialization_alias=serialization_alias,
            pydantic_type=pydantic_type,
        )

    def _is_type_repeatable(self) -> bool:
        """
        Heurist uses the code 0 to indicate that a record's detail (field) \
            can be repeated. Parse this information on the record structure \
            to determine a boolean indicating whether or not the detail is repeated.

        Returns:
            bool: Whether the detail can be repeated.
        """

        if self.rst_MaxValues == 0:
            return True
        else:
            return False

    def _compose_annotation(
        self, validation_alias: str, serialization_alias: str, pydantic_type: Any
    ) -> dict:
        """
        Using the Heurist information stored in the class instance's attributes, build \
            the Pydantic data field's alias and compose its annotation.

        Returns:
            dict: Key-value pair: Pydantic field's alias (key) - annotation (value)
        """

        if not validation_alias:
            raise ValueError()
        if not serialization_alias:
            raise ValueError()
        if not pydantic_type:
            raise ValueError()

        if self.rst_MaxValues == 0:
            default = []
        else:
            default = None

        return {
            validation_alias: (
                pydantic_type,
                Field(
                    # ID of the column's data type in Heurist
                    description=self.dty_ID,
                    # Formatted way to identify the Pydantic column
                    validation_alias=validation_alias,
                    # SQL-safe version of the column name
                    serialization_alias=serialization_alias,
                    # Default value to write in DuckDB column
                    default=default,
                ),
            )
        }



"""
CLI command for downloading all requested record types' data.
"""

from pathlib import Path

import duckdb
from heurist.api.connection import HeuristAPIConnection
from heurist.api.credentials import CredentialHandler
from heurist.log import log_summary
from heurist.log.constants import VALIDATION_LOG
from heurist.utils.constants import DEFAULT_RECORD_GROUPS
from heurist.workflows import extract_transform_load
from rich.columns import Columns
from rich.console import Console, Group
from rich.padding import Padding
from rich.panel import Panel


def load_command(
    credentials: CredentialHandler,
    duckdb_database_connection_path: Path | str,
    record_group: tuple = DEFAULT_RECORD_GROUPS,
    user: tuple = (),
    outdir: Path | None = None,
):
    # Run the ETL process
    if isinstance(duckdb_database_connection_path, Path):
        duckdb_database_connection_path = str(duckdb_database_connection_path)
    with (
        duckdb.connect(duckdb_database_connection_path) as conn,
        HeuristAPIConnection(
            db=credentials.get_database(),
            login=credentials.get_login(),
            password=credentials.get_password(),
        ) as client,
    ):
        extract_transform_load(
            client=client,
            duckdb_connection=conn,
            record_group_names=record_group,
            user=user,
        )

    # Show the results of the created DuckDB database
    with duckdb.connect(duckdb_database_connection_path, read_only=True) as new_conn:
        tables = [t[0] for t in new_conn.sql("show tables;").fetchall()]
        if VALIDATION_LOG.is_file():
            with open(VALIDATION_LOG) as f:
                log = f.readlines()
        else:
            log = []
        show_summary_in_console(tables=tables, log_lines=log)

        # If writing to CSV files, write only tables of record types
        if outdir:
            outdir = Path(outdir)
            outdir.mkdir(exist_ok=True)
            for table in tables:
                # Skip the schema tables
                if table in ["rtg", "rst", "rty", "dty", "trm"]:
                    continue
                fp = outdir.joinpath(f"{table}.csv")
                new_conn.table(table).sort("H-ID").write_csv(str(fp))


def show_summary_in_console(tables: list[str], log_lines: list):
    console = Console()
    t0 = Panel(
        Columns(tables, equal=True, expand=True),
        title="SQL Tables",
        subtitle="Saved in DuckDB database file.",
    )
    t1, t2 = log_summary(lines=log_lines)
    panel_group = Group(Padding(t0, 1), t1, Padding(t2, 1))
    console.print(panel_group)



from datetime import datetime
from typing import List, Literal, Optional

from heurist.models.structural.utils import split_ids
from pydantic import field_validator
from pydantic_xml import BaseXmlModel, element


class DTY(BaseXmlModel, tag="dty", search_mode="unordered"):
    """Dataclass to model one of the database's Detail Types. A Detail Type is the
        generic schema that defines the type of data one of a record's field.

    When possible, the attribute descriptions are taken from Heurist's source code.

    Attributes:
        dty_ID (int): Code for the detail type (field) - may vary between Heurist DBs.
        dty_Name (str): The canonical (standard) name of the detail type, used as \
            default in edit form.
        dty_Documentation (Optional[str]): Documentation of the detail type, what it \
            means, how defined.
        dty_Type (Literal['freetext','blocktext','integer','date','year','relmarker',\
            'boolean','enum','relationtype','resource','float','file','geo','separator',\
            'calculated','fieldsetmarker','urlinclude']): The value-type of this \
            detail type, what sort of data is stored.
        dty_HelpText (Optional[str]):The default help text displayed to the user under \
            the field.
        dty_ExtendedDescription (Optional[str]): Extended text describing this detail \
            type, for display in rollover.
        dty_EntryMask (Optional[str]): Data entry mask, use to control decimals on \
            numeric values, content of text fields etc.
        dty_Status (Literal["reserved","approved","pending","open"]): 'Reserved' for \
            the system, cannot be changed; 'Approved' for community standards; \
            'Pending' for work in progress; 'Open' for freely modifiable/personal \
            record types.
        dty_OriginatingDBID (int): Database where this detail type originated, 0 = \
            locally.
        dty_NameInOriginatingDB (Optional[str]): Name used in database where this \
            detail type originated.
        dty_IDInOriginatingDB (int): ID used in database where this detail type \
            originated.
        dty_DetailTypeGroupID (int): The general role of this detail allowing \
            differentiated lists of detail types.
        dty_OrderInGroup (int): The display order of DetailType within group, \
            alphabetic if equal values.
        dty_JsonTermIDTree (Optional[str]): Tree of Term IDs to show for this field \
            (display-only header terms set in HeaderTermIDs).
        dty_TermIDTreeNonSelectableIDs (List[Optional[int]]): Term IDs to use as \
            non-selectable headers for this field.
        dty_PtrTargetRectypeIDs (List[Optional[int]]): CSVlist of target Rectype IDs, \
            null = any.
        dty_FieldSetRectypeID (Optional[int]): For a FieldSetMarker, the record type \
            to be inserted as a fieldset.
        dty_ShowInLists (bool): Show this field type in pulldown lists etc. (always \
            visible in field management screen).
        dty_NonOwnerVisibility (Literal["hidden","viewable","public"]): Hidden = \
            visible only to owners, Viewable = any logged in user, Public = visible \
                to non-logged in viewers.
        dty_Modified (datetime): Date of last modification of this record, used to get \
            last updated date for table.
        dty_LocallyModified (bool): Flags a definition element which has been modified \
            relative to the original source.
        dty_SemanticReferenceURL (Optional[str]): URI to a full description or \
            ontological reference definition of the base field (optional).
    """

    dty_ID: int = element()
    dty_Name: str = element()
    dty_Documentation: Optional[str] = element(default=None)
    dty_Type: Literal[
        "freetext",
        "blocktext",
        "integer",
        "date",
        "year",
        "relmarker",
        "boolean",
        "enum",
        "relationtype",
        "resource",
        "float",
        "file",
        "geo",
        "separator",
        "calculated",
        "fieldsetmarker",
        "urlinclude",
    ] = element()
    dty_HelpText: Optional[str] = element(default=None)
    dty_ExtendedDescription: Optional[str] = element(default=None)
    dty_EntryMask: Optional[str] = element(default=None)
    dty_Status: Literal["reserved", "approved", "pending", "open"] = element()
    dty_OriginatingDBID: int = element()
    dty_NameInOriginatingDB: Optional[str] = element(default=None)
    dty_IDInOriginatingDB: int = element()
    dty_DetailTypeGroupID: int = element()
    dty_OrderInGroup: int = element()
    dty_JsonTermIDTree: Optional[str] = element(default=None)
    dty_TermIDTreeNonSelectableIDs: List[Optional[int]] = element(default=[])
    dty_PtrTargetRectypeIDs: List[Optional[int]] = element(default=[])
    dty_FieldSetRectypeID: Optional[int] = element(default=None)
    dty_ShowInLists: bool = element()
    dty_NonOwnerVisibility: Literal["hidden", "viewable", "public"] = element()
    dty_Modified: datetime = element()
    dty_LocallyModified: bool = element()
    dty_SemanticReferenceURL: Optional[str] = element(default=None)

    @field_validator("dty_TermIDTreeNonSelectableIDs", mode="before")
    @classmethod
    def validate_selectable_ids(cls, input_value: str | None) -> list:
        if input_value:
            return split_ids(input=input_value)
        else:
            return []

    @field_validator("dty_PtrTargetRectypeIDs", mode="before")
    @classmethod
    def validate_rectype_ids(cls, input_value: str | None) -> list:
        if input_value:
            return split_ids(input=input_value)
        else:
            return []



# Result of joining the database schema tables
METADATA = {
    "dty_ID": 1114,
    "rst_DisplayName": "is_expression_of",
    "dty_Type": "resource",
    "rst_MaxValues": 0,
}

# Result of a record's JSON export
DETAIL = [
    {
        "dty_ID": 1114,
        "value": {
            "id": "36",
            "type": "103",
            "title": "Dagobert",
            "hhash": None,
        },
        "fieldName": "is_expression_of",
        "fieldType": "resource",
        "conceptID": "",
    },
    {
        "dty_ID": 1114,
        "value": {
            "id": "79",
            "type": "103",
            "title": "Garin le Lorrain",
            "hhash": None,
        },
        "fieldName": "is_expression_of",
        "fieldType": "resource",
        "conceptID": "",
    },
]



from pathlib import Path

file_path_record_type_schema = Path(__file__).parent.joinpath(
    "selectRecordTypeSchema.sql"
)
file_path_record_type_metadata_by_group_type = Path(__file__).parent.joinpath(
    "joinRecordTypeIDNameByGroupType.sql"
)
file_path_fully_joined_record_type_metadata = Path(__file__).parent.joinpath(
    "joinRecordTypeMetadata.sql"
)


with open(file_path_record_type_schema) as f:
    RECORD_TYPE_SCHEMA = f.read()

with open(file_path_record_type_metadata_by_group_type) as f:
    RECORD_BY_GROUP_TYPE = f.read()

with open(file_path_fully_joined_record_type_metadata) as f:
    RECORD_TYPE_METADATA = f.read()



from datetime import datetime
from typing import Optional

from pydantic_xml import BaseXmlModel, element


class RTG(BaseXmlModel, tag="rtg", search_mode="unordered"):
    """Dataclass to model one of the database's Record Type Groups. A Record Type \
        Group categorizes the record types in the database.

    When possible, the attribute descriptions are taken from Heurist's source code.

    Attributes:
        rtg_ID (int): __Description__.
        rtg_Name (str): __Description__.
        rtg_Domain (str): __Description__.
        rtg_Description (Optional[str]): __Description__.
        rtg_Modified (datetime): __Description__.
    """

    rtg_ID: int = element()
    rtg_Name: str = element()
    rtg_Domain: str = element()
    rtg_Description: Optional[str] = element(default=None)
    rtg_Modified: datetime = element()



from datetime import datetime
from typing import Literal, Optional

from pydantic_xml import BaseXmlModel, element


class RTY(BaseXmlModel, tag="rty", search_mode="unordered"):
    """Dataclass to model one of the database's Record Types. A Record Type is the \
        schema for an entity in the database.

    When possible, the attribute descriptions are taken from Heurist's source code.

    Attributes:
        rty_ID (int): Record type code, widely used to reference record types, primary \
            key
        rty_Name (str): The name which is used to describe this record (object) type
        rty_OrderInGroup (int): Ordering within record type display groups for pulldowns
        rty_Description (str): Description of this record type
        rty_TitleMask (str): Mask to build a composite title by combining field values
        rty_CanonicalTitleMask (str): Version of the mask converted to detail codes \
            for processing
        rty_Plural (Optional[str]): Plural form of the record type name, manually \
            entered
        rty_Status (Literal["reserved", "approved", "pending", "open"]): Reserved \
            Heurist codes, approved/pending by ''Board'', and user additions
        rty_OriginatingDBID (Optional[int]): Database where this record type \
            originated, 0 = locally
        rty_NameInOriginatingDB: (Optional[str]) Name used in database where this \
            record type originated
        rty_IDInOriginatingDB (Optional[int]): ID in database where this record type \
            originated
        rty_NonOwnerVisibility (Literal["hidden", "viewable", "public", "pending"]): \
            Allows blanket restriction of visibility of a particular record type
        rty_ShowInLists (bool): Flags if record type is to be shown in end-user \
            interface, 1=yes
        rty_RecTypeGroupID (int): Record type group to which this record type belongs
        rty_RecTypeModelIDs (str): The model group(s) to which this rectype belongs, \
            comma sep. list
        rty_FlagAsFieldset (bool): 0 = full record type, 1 = Fieldset = set of fields \
            to include in other rectypes
        rty_ReferenceURL (Optional[str]): A semantic reference URI for, or a URL \
            describing, the record type
        rty_AlternativeRecEditor (Optional[str]): Name or URL of alternative record \
            editor function to be used for this rectype
        rty_Type (Literal["normal", "relationship", "dummy"]): Use to flag special \
            record types to trigger special functions
        rty_ShowURLOnEditForm (bool): Determines whether special URL field is shown at \
            the top of the edit form
        rty_ShowDescriptionOnEditForm (bool): Determines whether the record type \
            description field is shown at the top of the edit form
        rty_Modified (datetime): Date of last modification of this record, used to get \
            last updated date for table
        rty_LocallyModified (bool): Flags a definition element which has been modified \
            relative to the original source
    """

    rty_ID: int = element()
    rty_Name: str = element()
    rty_OrderInGroup: int = element()
    rty_Description: str = element()
    rty_TitleMask: str = element()
    rty_CanonicalTitleMask: Optional[str] = element(default=None)
    rty_Plural: Optional[str] = element(default=None)
    rty_Status: Literal["reserved", "approved", "pending", "open"] = element()
    rty_OriginatingDBID: Optional[int] = element(default=None)
    rty_NameInOriginatingDB: Optional[str] = element(default=None)
    rty_IDInOriginatingDB: Optional[int] = element(default=None)
    rty_NonOwnerVisibility: Literal["hidden", "viewable", "public", "pending"] = (
        element()
    )
    rty_ShowInLists: bool = element()
    rty_RecTypeGroupID: int = element()
    rty_RecTypeModelIDs: str = element(default=None)
    rty_FlagAsFieldset: bool = element(default=None)
    rty_ReferenceURL: Optional[str] = element(default=None)
    rty_AlternativeRecEditor: Optional[str] = element(default=None)
    rty_Type: Literal["normal", "relationship", "dummy"] = element()
    rty_ShowURLOnEditForm: bool = element()
    rty_ShowDescriptionOnEditForm: bool = element()
    rty_Modified: datetime = element()
    rty_LocallyModified: bool = element()



import unittest

from heurist.database.basedb import HeuristDatabase

from mock_data import DB_STRUCTURE_XML


class DuckBaseTest(unittest.TestCase):
    def test(self):
        """Test should show that the 5 basic data models from the HML XML
        were converted to SQL tables."""

        self.db = HeuristDatabase(hml_xml=DB_STRUCTURE_XML)
        rel = self.db.conn.sql("show tables")
        self.assertEqual(len(rel), 5)


if __name__ == "__main__":
    unittest.main()



from heurist.models.structural.rtg import RTG
from pydantic_xml import BaseXmlModel, element


class RecTypeGroups(BaseXmlModel):
    """Dataclass for modeling all of the database structure's Record Type Groups.

    Attributes:
        rtg (list): list of instantiated dataclasses that model all of the database's
            Record Type Groups.

    Examples:
        >>> from mock_data import DB_STRUCTURE_XML
        >>> from heurist.models.structural import HMLStructure
        >>>
        >>>
        >>> # Parse structure
        >>> xml = DB_STRUCTURE_XML
        >>> hml = HMLStructure.from_xml(xml)
        >>>
        >>> # Test class
        >>> first_record_type = hml.RecTypeGroups.rtg[0]
        >>> first_record_type.rtg_ID
        4
    """

    rtg: list[RTG] = element()



from datetime import datetime
from typing import Literal, Optional

from pydantic_xml import BaseXmlModel, element


class RST(BaseXmlModel, tag="rst", search_mode="unordered"):
    """Dataclass to model one of the database's Record Structures. Record Structures
        are the fields of a Record Type.

    When possible, the attribute descriptions are taken from Heurist's source code.

    Attributes:
        rst_ID (int): Primary key for the record structures table
        rst_RecTypeID (int): The record type to which this detail is allocated, \
            0 = all rectypes
        rst_DetailTypeID (int): Detail type for this field or, if MSB set, \
            FieldSet code + 32767
        rst_DisplayName (str): Display name for this dtl type in this rectype, \
            autofill with dty_Name
        rst_DisplayHelpText (Optional[str]): The user help text to be displayed for \
            this detail type for this record type
        rst_DisplayExtendedDescription (Optional[str]): The rollover text to be \
            displayed for this detail type for this record type
        rst_DisplayOrder (int): A sort order for display of this detail type in the \
            record edit form
        rst_DisplayWidth (int): The field width displayed for this detail type in \
            this record type
        rst_DisplayHeight (int): The field height for this detail type in this record \
            type, only relevant for memo fields
        rst_DefaultValue (Optional[str]): The default value for this detail type for \
            this record type
        rst_RecordMatchOrder (int): Indicates order of significance in detecting \
            duplicate records, 1 = highest
        rst_CalcFunctionID (Optional[int]): FK to table of function specifications for \
            calculating string values
        rst_CalcFieldMask (Optional[str]): A mask string along the lines of the title \
            mask allowing a composite field to be generated from other fields in the \
            record
        rst_RequirementType (Literal["required", "recommended", "optional", \
            "forbidden"]):
        rst_NonOwnerVisibility (Literal["hidden", "viewable", "public", "pending"]): \
            Allows restriction of visibility of a particular field in a specified \
            record type
        rst_Status (Literal["reserved", "approved", "pending", "open"]): Reserved \
            Heurist codes, approved/pending by ''Board'', and user additions
        rst_MayModify (Literal["locked", "discouraged", "open"]): Extent to which \
            detail may be modified within this record structure
        rst_OriginatingDBID (int): Database where this record structure element \
            originated, 0 = locally
        rst_IDInOriginatingDB (Optional[int]): ID used in database where this record \
            structure element originated
        rst_MaxValues (int): Maximum number of values per record for this detail, 1 - \
            single, >1 limited, NULL or 0 = no limit
        rst_MinValues (int): If required, minimum number of values per record for this \
            detail
        rst_InitialRepeats (int): Number of repeat values to be displayed for this \
            field when a new record is first displayed
        rst_DisplayDetailTypeGroupID (Optional[int]): If set, places detail in \
            specified group instead of according to dty_DetailTypeGroup
        rst_FilteredJsonTermIDTree (Optional[str]): JSON encoded tree of allowed \
            terms, subset of those defined in defDetailType. This field is no \
            longer used
        rst_PtrFilteredIDs (Optional[str]): Allowed Rectypes (CSV) within list defined \
            by defDetailType (for pointer details) This field is no longer used
        rst_CreateChildIfRecPtr (bool): For pointer fields, flags that new records \
            created from this field should be marked as children of the creating record
        rst_PointerMode (Literal["dropdown_add", "dropdown", "addorbrowse", "addonly", \
            "browseonly"]): When adding record pointer values, default or null = show \
            both add and browse, otherwise only allow add or only allow \
            browse-for-existing
        rst_PointerBrowseFilter (Optional[str]): When adding record pointer values, \
            defines a Heurist filter to restrict the list of target records browsed
        rst_OrderForThumbnailGeneration (Optional[str]): Priority order of fields to \
            use in generating thumbnail, null = do not use
        rst_TermIDTreeNonSelectableIDs (Optional[str]): Term IDs to use as \
            non-selectable headers for this field
        rst_ShowDetailCertainty (bool): When editing the field, allow editng of the \
            dtl_Certainty value (off by default)
        rst_ShowDetailAnnotation (bool): When editing the field, allow editng of the \
            dtl_Annotation value (off by default)
        rst_NumericLargestValueUsed (Optional[int]): For numeric fields, Null = \
            no auto increment, 0 or more indicates largest value used so far. \
            Set to 0 to switch on incrementing
        rst_EntryMask (Optional[str]): Data entry mask, use to control decimals on \
            numeric values, content of text fields etc. for this record type - future \
            implementation Aug 2017
        rst_Modified (datetime): Date of last modification of this record, used to get \
            last updated date for table
        rst_LocallyModified (int): Flags a definition element which has been modified \
            relative to the original source
        rst_SemanticReferenceURL (Optional[str]): The URI to a semantic definition or \
            web page describing this field used within this record type
        rst_TermsAsButtons (bool): If 1, term list fields are represented as buttons \
            (if single value) or checkboxes (if repeat values)
    """

    rst_ID: int = element()
    rst_RecTypeID: int = element()
    rst_DetailTypeID: int = element()
    rst_DisplayName: str = element()
    rst_DisplayHelpText: Optional[str] = element(default=None)
    rst_DisplayExtendedDescription: Optional[str] = element(default=None)
    rst_DisplayOrder: int = element()
    rst_DisplayWidth: int = element()
    rst_DisplayHeight: int = element()
    rst_DefaultValue: Optional[str] = element(default=None)
    rst_RecordMatchOrder: int = element()
    rst_CalcFunctionID: Optional[int] = element(default=None)
    rst_CalcFieldMask: Optional[str] = element(default=None)
    rst_RequirementType: Literal["required", "recommended", "optional", "forbidden"] = (
        element()
    )
    rst_NonOwnerVisibility: Literal["hidden", "viewable", "public", "pending"] = (
        element()
    )
    rst_Status: Literal["reserved", "approved", "pending", "open"] = element()
    rst_MayModify: Literal["locked", "discouraged", "open"] = element()
    rst_OriginatingDBID: int = element()
    rst_IDInOriginatingDB: Optional[int] = element(default=None)
    rst_MaxValues: int = element()
    rst_MinValues: int = element()
    rst_InitialRepeats: int = element()
    rst_DisplayDetailTypeGroupID: Optional[int] = element(default=None)
    rst_FilteredJsonTermIDTree: Optional[str] = element(default=None)
    rst_PtrFilteredIDs: Optional[str] = element(default=None)
    rst_CreateChildIfRecPtr: bool = element()
    rst_PointerMode: Literal[
        "dropdown_add", "dropdown", "addorbrowse", "addonly", "browseonly"
    ] = element()
    rst_PointerBrowseFilter: Optional[str] = element(default=None)
    rst_OrderForThumbnailGeneration: Optional[str] = element(default=None)
    rst_TermIDTreeNonSelectableIDs: Optional[str] = element(default=None)
    rst_ShowDetailCertainty: bool = element()
    rst_ShowDetailAnnotation: bool = element()
    rst_NumericLargestValueUsed: Optional[int] = element(default=None)
    rst_EntryMask: Optional[str] = element(default=None)
    rst_Modified: datetime = element()
    rst_LocallyModified: int = element()
    rst_SemanticReferenceURL: Optional[str] = element(default=None)
    rst_TermsAsButtons: bool = element()



# Result of joining the database schema tables
METADATA = {
    "dty_ID": 1114,
    "rst_DisplayName": "is_expression_of",
    "dty_Type": "resource",
    "rst_MaxValues": 1,
}

# Result of a record's JSON export
DETAIL = {
    "dty_ID": 1114,
    "value": {"id": "36", "type": "103", "title": "Dagobert", "hhash": None},
    "fieldName": "is_expression_of",
    "fieldType": "resource",
    "conceptID": "",
}



import re

import duckdb

KEYWORDS = duckdb.sql("select * from duckdb_keywords()").fetchall()


class SafeSQLName:
    def __init__(self) -> None:
        self.reserved = [t[0] for t in KEYWORDS if t[1] == "reserved"]
        self.unreserved = [t[0] for t in KEYWORDS if t[1] == "unreserved"]
        self.column_name = [t[0] for t in KEYWORDS if t[1] == "column_name"]
        self.type_function = [t[0] for t in KEYWORDS if t[1] == "type_function"]
        self.all_keywords = [t[0] for t in KEYWORDS]

    @classmethod
    def remove_characters(cls, s: str) -> str:
        """Simplify and remove undesirable characters from a string.

        Examples:
        >>> s = "Author or Creator (Person, Organization)"
        >>> SafeSQLName.remove_characters(s)
        'Author or Creator'

        >>> s = "Status_trad_freetext"
        >>> SafeSQLName.remove_characters(s)
        'Status_trad_freetext'

        Args:
            s (str): Input string.

        Returns:
            str: Cleaned string.
        """

        # Remove parentheses
        s = re.sub(r"\(.+\)", "", s)
        # Remove non-letters
        s = re.sub(r"\W", " ", s)
        # Remove backslashes
        s = re.sub(r"/", " ", s)
        # Remove double spaces
        s = re.sub(r"\s+", " ", s)
        # Remove double underscores
        s = re.sub(r"_+", "_", s)
        # Trim underscores
        s = s.strip()
        return s

    @classmethod
    def to_pascal_case(cls, text: str) -> str:
        text_string = text.replace("-", " ").replace("_", " ")
        words = text_string.split()
        if len(text) == 0:
            return text
        capitalized_words = ["".join(w[0].capitalize() + w[1:] for w in words)]
        return "".join(capitalized_words)

    def create_column_name(self, field_name: str, field_type: str) -> str:
        """
        Create an SQL-safe column name for the Pydantic data field.

        Args:
            field_name (str): Displayed name of the field (detail) in Heurist.
            field_type (str): Heurist type of the field (detail).

        Returns:
            str: SQL-safe column name.
        """

        simplified_name = self.remove_characters(field_name)
        if field_type == "resource":
            final_name = f"{simplified_name} H-ID"
        elif simplified_name.lower() in self.all_keywords:
            final_name = f"{simplified_name}_COLUMN"
        else:
            final_name = simplified_name
        return final_name

    def create_table_name(self, record_name: str) -> str:
        """
        Create SQL-safe table name for the record's data model.

        Examples:
        >>> heurist_name = "Sequence"
        >>> SafeSQLName().create_table_name(heurist_name)
        'SequenceTable'

        Args:
            record_name (str): Name of the Heurist record type.

        Returns:
            str: SQL-safe name for the record type's table.
        """

        camel_case_name = self.to_pascal_case(record_name)
        if camel_case_name.lower() in self.all_keywords:
            final_name = f"{camel_case_name}Table"
        else:
            final_name = camel_case_name
        return final_name



"""
CLI command for downloading details about a Heurist database schema.
"""

from heurist.api.credentials import CredentialHandler
from heurist.schema import export_schema


def schema_command(
    credentials: CredentialHandler,
    record_group: tuple,
    outdir: str,
    output_type: str,
    debugging: bool = False,
):
    export_schema(
        db_name=credentials.get_database(),
        login=credentials.get_login(),
        password=credentials.get_password(),
        outdir=outdir,
        debugging=debugging,
        output_type=output_type,
        record_group=record_group
    )



import json
from pathlib import Path

DIR = Path(__file__).parent

DB_STRUCTURE_XML_FILE = DIR.joinpath("hml.xml")
RECORD_JSON_FILE = DIR.joinpath("records.json")
FUZZY_DATE_JSON_FILE = DIR.joinpath("fuzzydate_record.json")


with open(DB_STRUCTURE_XML_FILE, "rb") as f:
    DB_STRUCTURE_XML = f.read()


with open(RECORD_JSON_FILE) as f:
    r = f.read()
    RECORD_JSON = json.loads(r)


with open(FUZZY_DATE_JSON_FILE) as f:
    r = f.read()
    FUZZY_DATE_RECORD_JSON = json.loads(r)



import re
from dataclasses import dataclass


@dataclass
class LogDetail:
    time: str
    level: str
    recType: int
    recID: int
    rule: str
    problem: str

    @classmethod
    def load_lines(cls, *block_lines) -> "LogDetail":
        l1, l2, l3, l4, l5 = block_lines
        for indented_line in [l2, l3, l4, l5]:
            assert indented_line.startswith("\t")
        return LogDetail(
            time=cls.parse_time(l1),
            level=cls.parse_level(l1),
            recType=cls.parse_number(l2),
            recID=cls.parse_number(l3),
            rule=l4.removeprefix("\t").strip(),
            problem=l5.removeprefix("\t").strip(),
        )

    @staticmethod
    def parse_number(line) -> int:
        parts = line.split()
        suffix: str = parts[-1]
        number = suffix.removesuffix("]")
        return int(number)

    @staticmethod
    def parse_time(l1: str) -> str:
        parts = l1.split(" - ")
        return parts[0].strip()

    @staticmethod
    def parse_level(l1: str) -> str:
        return re.search(r"[A-Z]+", l1).group(0).strip()



import unittest

from heurist.api.credentials import CredentialHandler
from heurist.schema.export import get_database_schema


class SchemaTest(unittest.TestCase):
    def setUp(self):
        try:
            self.credentials = CredentialHandler()
        except SystemExit:
            self.skipTest(
                "Connection could not be established.\nCannot test client without \
                    database connection."
            )
        self.debugging = True

    def test(self):
        db = get_database_schema(
            record_groups=["My record types"],
            credentials=self.credentials,
            debugging=True,
        )
        actual = len(db.pydantic_models)
        self.assertGreater(actual, 0)


if __name__ == "__main__":
    unittest.main()



"""
CLI commands for extracting, transforming, and loading remote Heurist data.
"""

import importlib.metadata

import click
from heurist import PACKAGE_NAME
from heurist.api.credentials import CredentialHandler
from heurist.api.exceptions import MissingParameterException
from heurist.cli.load import load_command
from heurist.cli.records import rty_command
from heurist.cli.schema import schema_command
from heurist.utils.constants import DEFAULT_RECORD_GROUPS
from rich.console import Console

# This name must match the package name ('name' kwarg) in the TOML file.
__identifier__ = importlib.metadata.version(PACKAGE_NAME)


# =========================== #
#     Main cli group
# =========================== #
@click.group(help="Group CLI command for connecting to the Heurist DB")
@click.version_option(__identifier__)
@click.option(
    "-d",
    "--database",
    type=click.STRING,
    help="Name of the Heurist database",
)
@click.option(
    "-l",
    "--login",
    type=click.STRING,
    help="Login name for the database user",
)
@click.option(
    "-p",
    "--password",
    type=click.STRING,
    help="Password for the database user",
)
@click.option(
    "--debugging",
    required=False,
    default=False,
    is_flag=True,
    help="Whether to run in debug mode, default false.",
)
@click.pass_context
def cli(ctx, database, login, password, debugging):
    ctx.ensure_object(dict)
    ctx.obj["DEBUGGING"] = debugging
    try:
        ctx.obj["CREDENTIALS"] = CredentialHandler(
            database_name=database,
            login=login,
            password=password,
        )
    except MissingParameterException:
        c = Console()
        c.print(
            "Login informaiton is missing."
            "Please provide your credentials when prompted."
            "\nTo quit, press Ctrl+C then Enter."
        )
        _database = click.prompt("Heurist database name")
        _login = click.prompt("Heurist user login")
        _password = click.prompt("Heurist login password")
        c.print("Retrying the connection...")
        ctx.obj["CREDENTIALS"] = CredentialHandler(
            database_name=_database,
            login=_login,
            password=_password,
        )
        c.print("Success!", style="green")


# =========================== #
#     'record' command
# =========================== #
@cli.command("record", help="Get a JSON export of a certain record type.")
@click.option(
    "-t",
    "--record-type",
    help="The ID fo the record type",
    type=click.INT,
    required=True,
)
@click.option(
    "-o",
    "--outfile",
    help="JSON file path.",
    type=click.Path(file_okay=True, writable=True),
    required=False,
)
@click.pass_obj
def records(ctx, record_type, outfile):
    credentials = ctx["CREDENTIALS"]
    rty_command(credentials, record_type, outfile)


# =========================== #
#     'schema' command
# =========================== #
@cli.command(
    "schema",
    help="Generate documentation about the database schema.",
)
@click.option(
    "-t",
    "--output-type",
    required=True,
    type=click.Choice(["csv", "json"], case_sensitive=False),
    help="Data format in which the schema will be described. \
    csv = 1 CSV file for each record type. json = 1 file that \
    lists all records together",
)
@click.option(
    "-r",
    "--record-group",
    required=False,
    type=click.STRING,
    multiple=True,
    default=DEFAULT_RECORD_GROUPS,
    show_default=True,
    help="Group name of the record types to be described. \
        Can be declared multiple times for multiple groups.",
)
@click.option(
    "-o",
    "--outdir",
    required=False,
    type=click.Path(file_okay=False, dir_okay=True),
    help="Path to the directory in which the files will be written. \
        Defaults to name of the database + '_schema'.",
)
@click.pass_obj
def doc(ctx, record_group, outdir, output_type):
    # Get context variables
    credentials = ctx["CREDENTIALS"]
    debugging = ctx["DEBUGGING"]

    # Run the doc command
    schema_command(
        credentials=credentials,
        record_group=record_group,
        outdir=outdir,
        output_type=output_type,
        debugging=debugging,
    )


# =========================== #
#     'download' command
# =========================== #
@cli.command(
    "download",
    help="Export data of records of 1 or more record group types.",
)
@click.option(
    "-r",
    "--record-group",
    required=False,
    type=click.STRING,
    multiple=True,
    default=DEFAULT_RECORD_GROUPS,
    help="Record group of the entities whose data is exported. \
        Default: 'My record types'.",
)
@click.option(
    "-u",
    "--user",
    required=False,
    type=click.INT,
    multiple=True,
    help="User or users who created the records to be exported. \
        Default: all users' records.",
)
@click.option(
    "-f",
    "--filepath",
    required=True,
    type=click.Path(
        file_okay=True,
        dir_okay=False,
    ),
    help="Path to the DuckDB database file in which the data will be written.",
)
@click.option(
    "-o",
    "--outdir",
    required=False,
    type=click.Path(
        file_okay=False,
        dir_okay=True,
    ),
    help="Directory in which CSV files of the dumped tabular data \
        will be written.",
)
@click.pass_obj
def load(ctx, filepath, record_group, user, outdir):
    # Get context variable
    credentials = ctx["CREDENTIALS"]
    testing = ctx["DEBUGGING"]

    # Run the dump command
    if not testing:
        load_command(
            credentials=credentials,
            duckdb_database_connection_path=filepath,
            record_group=record_group,
            user=user,
            outdir=outdir,
        )
    else:
        print(
            "\nCannot run 'dump' command in debugging mode.\
            \nClient must connect to a remote Heurist database.\n"
        )
        print("Exiting.")
        exit()


if __name__ == "__main__":
    cli()



import pandas as pd
from duckdb import DuckDBPyConnection, DuckDBPyRelation
from heurist.database.basedb import HeuristDatabase
from heurist.models.dynamic import HeuristRecord
from heurist.sql import RECORD_BY_GROUP_TYPE, RECORD_TYPE_METADATA
from heurist.validators.record_validator import RecordValidator
from heurist.utils.constants import DEFAULT_RECORD_GROUPS


class TransformedDatabase(HeuristDatabase):
    """Class for building and populating SQL tables with data collected and \
    transformed from remote Heurist DB.
    """

    def __init__(
        self,
        hml_xml: bytes,
        conn: DuckDBPyConnection | None = None,
        db: str | None = ":memory:",
        record_type_groups: tuple = DEFAULT_RECORD_GROUPS,
    ) -> None:
        super().__init__(hml_xml, conn, db)

        self.conn.execute("SET GLOBAL pandas_analyze_sample=100000")

        # Create an empty index of targeted record types' Pydantic models
        self.pydantic_models = {}

        # Joining together the Heurist database's structural tables, construct an SQL
        # statement that selects the ID and name of the record types that belong to one
        # of the targeted record type groups
        condition = "\nWHERE rtg.rtg_Name like '{}'".format(record_type_groups[0])
        if len(record_type_groups) > 1:
            for rtg in record_type_groups[1:]:
                condition += " OR rtg.rtg_Name like '{}'".format(rtg)
        query = RECORD_BY_GROUP_TYPE + condition

        # Iterate through each targeted record type's ID and name
        for rty_ID, rty_Name in self.conn.sql(query).fetchall():
            # Using the ID, select the metadata of a record type's data fields (details)
            rel = self.conn.sql(query=RECORD_TYPE_METADATA, params=[rty_ID])
            # Using this metadata, create a dynamic Pydantic model for the record type
            data_field_metadata = rel.pl().to_dicts()
            model = HeuristRecord(
                rty_ID=rty_ID,
                rty_Name=rty_Name,
                detail_metadata=data_field_metadata,
            )
            # Add the dynamic Pydantic model to the index of models
            self.pydantic_models.update({rty_ID: model})

    def insert_records(
        self, record_type_id: int, records: list[dict]
    ) -> DuckDBPyRelation | None:
        # From the index of Pydantic models, get this record type's
        # dynamically-created Pydantic model.
        dynamic_model = self.pydantic_models[record_type_id].model
        table_name = self.pydantic_models[record_type_id].table_name

        # Prepare a list in which to store dictionaries of the validated record data.
        model_dict_sequence = []

        # Using the dynamically-created Pyandtic model, validate the metadata of
        # all the records of this type.
        validator = RecordValidator(
            pydantic_model=dynamic_model,
            records=records,
            rty_ID=record_type_id,
        )
        for model in validator:
            # Dump the validated record's data model to a dictionary.
            model_dict = model.model_dump(by_alias=True)
            # Add the dictionary representation of the validated data to the sequence.
            model_dict_sequence.append(model_dict)

        # If no records of this type have been created yet, skip it.
        if len(model_dict_sequence) == 0:
            return

        # Transform the sequence of dictionaries into a Pandas dataframe
        try:
            df = pd.DataFrame(model_dict_sequence)
            df = df.convert_dtypes(dtype_backend="numpy_nullable")
            assert df.shape[1] > 0
        except Exception as e:
            from pprint import pprint

            pprint(model_dict_sequence)
            print(df)
            print(records)
            print(table_name)
            raise e

        # Delete any existing table for this record type.
        self.delete_existing_table(table_name=table_name)

        # From the dataframe, build a new table for the record type.
        sql = f"""CREATE TABLE {table_name} AS FROM df"""
        self.conn.sql(sql)
        return self.conn.table(table_name=table_name)



"""Class for converting a record's detail before the Pydantic model validation."""

from heurist.models.dynamic.date import TemporalObject
from heurist.models.dynamic.type import FieldType


class DetailValidator:
    """
    In Heurist, a record's "detail" is what is more commonly known as an attribute, \
        dimension, or a data field.

    This class features methods to extract the key value from Heurist's JSON \
        formatting for all data types in Heurist's system.
    """

    direct_values = ["freetext", "blocktext", "integer", "boolean", "float"]

    @classmethod
    def validate_file(cls, detail: dict) -> str:
        """
        Extract the value of a file field.

        Args:
            detail (dict): Record's detail.

        Returns:
            str: Value of record's detail.
        """

        return detail.get("value", {}).get("file", {}).get("ulf_ExternalFileReference")

    @classmethod
    def validate_enum(cls, detail: dict) -> str:
        """
        Extract the value of an enum field.

        Args:
            detail (dict): Record's detail.

        Returns:
            str: Value of record's detail.
        """

        return detail["termLabel"]

    @classmethod
    def validate_geo(cls, detail: dict) -> str:
        """
        Extract the value of a geo field.

        Examples:
            >>> from mock_data.geo.single import DETAIL_POINT
            >>> DetailValidator.convert(DETAIL_POINT)
            'POINT(2.19726563 48.57478991)'

        Args:
            detail (dict): Record's detail.

        Returns:
            str: Value of record's detail.
        """

        geo = detail["value"]["geo"]
        if geo["type"] == "p" or geo["type"] == "pl":
            return geo["wkt"]

    @classmethod
    def validate_date(cls, detail: dict) -> dict:
        """
        Build the variable date value into a structured dictionary.

        Examples:
            >>> # Test temporal object
            >>> from mock_data.date.compound_single import DETAIL
            >>> value = DetailValidator.convert(DETAIL)
            >>> value['start']['earliest']
            datetime.datetime(1180, 1, 1, 0, 0)

            >>> # Test direct date value
            >>> from mock_data.date.simple_single import DETAIL
            >>> value = DetailValidator.convert(DETAIL)
            >>> value['value']
            datetime.datetime(2024, 3, 19, 0, 0)

        Args:
            detail (dict): Record's detail.

        Returns:
            dict: Structured metadata for a Heurist date object.
        """

        if isinstance(detail.get("value"), dict):
            model = TemporalObject.model_validate(detail["value"])
        else:
            model = TemporalObject.model_validate(detail)
        return model.model_dump(by_alias=True)

    @classmethod
    def validate_resource(cls, detail: dict) -> int:
        """
        Extract the value of a resource (foreign key) field.

        Args:
            detail (dict): Record's detail.

        Returns:
            int: Heurist ID of the referenced record.
        """

        return int(detail["value"]["id"])

    @classmethod
    def convert(cls, detail: dict) -> str | int | list | dict | None:
        """
        Based on the data type, convert the record's nested detail to a flat value.

        Args:
            detail (dict): One of the record's details (data fields).

        Returns:
            str | int | list | dict | None: Flattened value of the data field.
        """

        fieldtype = FieldType.from_detail(detail)

        if any(ft in fieldtype for ft in cls.direct_values):
            return detail["value"]

        elif fieldtype == "date":
            return cls.validate_date(detail)

        elif fieldtype == "enum":
            return cls.validate_enum(detail)

        elif fieldtype == "file":
            return cls.validate_file(detail)

        elif fieldtype == "geo":
            return cls.validate_geo(detail)

        elif fieldtype == "resource":
            return cls.validate_resource(detail)



import csv
from pathlib import Path

import click
from heurist.log import LogDetail, yield_log_blocks
from heurist.log.constants import VALIDATION_LOG

log_detail_fieldnames = list(LogDetail.__annotations__.keys())


@click.command()
@click.option("-l", "--log-file", required=None, default=VALIDATION_LOG)
@click.option("-o", "--outfile", required=None, default="invalid_records.csv")
def cli(log_file, outfile):
    logfile = Path(log_file)
    if not logfile.is_file():
        raise FileNotFoundError(log_file)
    with open(logfile) as f, open(outfile, "w") as of:
        writer = csv.DictWriter(of, fieldnames=log_detail_fieldnames)
        writer.writeheader()
        lines = f.readlines()
        for block in yield_log_blocks(lines):
            writer.writerow(block.__dict__)


if __name__ == "__main__":
    cli()



from .constants import MAX_RETRY


class APIException(Exception):
    """Problem calling Heurist API."""


class AuthenticationError(Exception):
    """Error raised when unable to authenticate Heurist login."""

    def __init__(self, message):
        self.message = """Authentication Error.
\tFailed to authenticate Heurist user login.
"""
        super().__init__(self.message)


class MissingParameterException(Exception):
    """Exception raised for a missing parameter."""

    def __init__(self, parameter: str, env_file: str):
        self.message = f"""MissingParameter Exception.
\tMissing the variable '{parameter}'.
\tTried looking in the env file '{env_file}'.
"""
        super().__init__(self.message)


class ReadTimeout(Exception):
    """Exception raised because the data returned by the Heurist \
        server took too long to receive.
    """

    def __init__(self, url: str, timeout: int):
        message = f"""ReadTimeout Error.
\tOn all {MAX_RETRY} tries, Heurist's server took too long (> {timeout} seconds) to \
send data from the following URL:
{url}
Solutions:
\t1. Try running the command again and hope the server / your internet is faster.
\t2. Set the READTIMEOUT environment variable immediately before the command and run it\
 again, i.e. 'READTIMEOUT=20 heurist download'.
"""
        self.message = message
        super().__init__(self.message)



import unittest
from pathlib import Path

from heurist.log import yield_log_blocks, LogDetail


LOG_FILE = Path(__file__).parent.joinpath("mock_log.txt")

LINES = [
    "2025-05-26 18:50 - WARNING - ",
    "	[rec_Type 104]",
    "	[rec_ID 834]",
    "	The detail 'note' is limited to a maximum of 1 values.",
    "	Count of values = 3.",
    "2025-05-26 18:50 - WARNING - ",
    "	[rec_Type 104]",
    "	[rec_ID 834]",
    "	The detail 'contam' is limited to a maximum of 1 values.",
    "	Count of values = 2.",
]


class TestLog(unittest.TestCase):
    def assess_log_block(self, log: LogDetail):
        self.assertEqual(log.level, "WARNING")
        self.assertIsNotNone(log.rule)
        self.assertIsNotNone(log.problem)
        self.assertGreater(log.recID, 100)
        self.assertGreater(log.recType, 99)

    def test_parsed_lines(self):
        for log in yield_log_blocks(LINES):
            self.assess_log_block(log)

    def test_log_file(self):
        print(LOG_FILE)
        with open(LOG_FILE) as f:
            lines = f.readlines()
            for log in yield_log_blocks(lines):
                self.assess_log_block(log)


if __name__ == "__main__":
    unittest.main()



from typing import List, Optional

from pydantic import BaseModel
from pydantic.functional_validators import BeforeValidator
from typing_extensions import Annotated


def convert_vocab_map_to_list(vocab_map: dict | None) -> list[dict]:
    """
    Convert map generated in the SQL query to a list of vocab dictionary objects.

    See the SQL query in the sql/ directory. The relevant selection is:
        map(list(trm_Label), list(
            {
                "description": trm_Description,
                "url": trm_SemanticReferenceURL,
                "id": trm_ID
            })
        ) AS terms

    Examples:
        >>> vocab0 = {'perg': {'description': 'Parchment', 'url': None, 'id': 9782}}
        >>> vocab1 = {'chart': {'description': 'Paper', 'url': None, 'id': 9783}}
        >>> vocab2 = {'mixed': {'description': None, 'url': None, 'id': 9785}}
        >>> map = vocab0 | vocab1 | vocab2
        >>> terms = convert_vocab_map_to_list(map)
        >>> len(terms)
        3
        >>> terms[0]
        {'label': 'perg', 'description': 'Parchment', 'url': None, 'id': 9782}

    Args:
        vocab_map (dict | None): Map created in SQL query from aggregation function.

    Returns:
        list[dict]: List of vocabulary term metadata in dictionary objects.
    """

    vocab_terms = []
    if vocab_map:
        for k, v in vocab_map.items():
            nd = {"label": k} | v
            vocab_terms.append(nd)
    return vocab_terms


VocabTerms = Annotated[list, BeforeValidator(convert_vocab_map_to_list)]


class DTY(BaseModel):
    rst_DisplayName: str
    rst_DisplayHelpText: str
    dty_ID: int
    dty_Type: str
    dty_PtrTargetRectypeIDs: Optional[List[int]]
    dty_SemanticReferenceURL: Optional[str]
    trm_TreeID: Optional[int]
    trm_Label: Optional[str]
    trm_Description: Optional[str]
    rst_RequirementType: str
    rst_MaxValues: int
    vocabTerms: Optional[VocabTerms]


class RTY(BaseModel):
    rty_ID: int
    rty_Name: str
    rty_Description: str
    rty_TitleMask: str
    rty_ReferenceURL: Optional[str]



from heurist.models.structural.DetailTypes import (
    DetailTypes as DetailTypesModel,
)
from heurist.models.structural.RecStructure import (
    RecStructure as RecStructureModel,
)
from heurist.models.structural.RecTypeGroups import (
    RecTypeGroups as RecTypeGroupsModel,
)
from heurist.models.structural.RecTypes import RecTypes as RecTypesModel
from heurist.models.structural.Terms import Terms as TermsModel
from pydantic_xml import BaseXmlModel, element


class HMLStructure(BaseXmlModel, tag="hml_structure", search_mode="unordered"):
    """Parent dataclass for modeling the entire Heurist database structure.

    Attributes:
        detail_types (DetailTypes): model for data nested in the DetailTypes tag.
        record_structures (RecStructure): model for data nested in the RecStructure tag.
        record_types (RecTypes): model for data nested in the RecTypes tag.

    Examples:
        >>> from mock_data import DB_STRUCTURE_XML
        >>>
        >>>
        >>> # Parse structure
        >>> xml = DB_STRUCTURE_XML
        >>> hml = HMLStructure.from_xml(xml)
    """

    DetailTypes: DetailTypesModel = element(tag="DetailTypes")
    RecStructure: RecStructureModel = element(tag="RecStructure")
    RecTypes: RecTypesModel = element(tag="RecTypes")
    RecTypeGroups: RecTypeGroupsModel = element(tag="RecTypeGroups")
    Terms: TermsModel = element(tag="Terms")



"""
Exceptions for classes that convert / transform Heurist data.
"""


class RepeatedValueInSingularDetailType(Exception):
    """The detail type is limited to a maximum of 1 values
    but the record has more than 1 value for this detail."""

    description = """
\t[rec_Type {typeID}]
\t[rec_ID {recID}]
\tThe detail '{fieldName}' is limited to a maximum of 1 values.
\tCount of values = {valueCount}."""

    def __init__(self, type_id: int, record_id: int, field_name: str, value_count: int):
        self.message = self.description.format(
            typeID=type_id,
            recID=record_id,
            fieldName=field_name,
            valueCount=value_count,
        )
        super().__init__(self.message)


class DateNotEnteredAsDateObject(Exception):
    """The date field was not entered as a constructed Heurist date object."""

    description = """The date field was not entered as a compound Heurist date \
object.\n\tEntered value = {}"""

    def __init__(self, value: int | str | float):
        self.message = self.description.format(value)
        super().__init__(self.message)



from datetime import datetime

from .constants import MAX_RETRY


def log_attempt_number(retry_state) -> None:
    """Simple logger for tenacity retry.

    Args:
        retry_state (tenacity.RetryCallState): Retry result from tenacity.
    """

    # Get the attempt number from the RetryCallState
    attempt_number = retry_state.attempt_number

    # Compose the console message
    time = datetime.now().strftime("%H:%M:%S")
    message = (
        f"[{time}] ReadTimeout error when receiving data from the Heurist API. "
        f"Retrying {attempt_number} / {MAX_RETRY} times..."
    )

    # Print the error message
    print(message)



# Result of joining the database schema tables
METADATA = {
    "dty_ID": 1111,
    "rst_DisplayName": "date_of_creation",
    "dty_Type": "date",
    "rst_MaxValues": 1,
}

# Result of a record's JSON export
DETAIL = {
    "dty_ID": 1111,
    "value": {
        "start": {
            "earliest": "1180",
            "latest": "1231",
            "profile": "1",
        },
        "end": {
            "latest": "1250",
            "earliest": "1246",
            "profile": "3",
        },
        "determination": "2",
        "estMinDate": 1180,
        "estMaxDate": 1250.1231,
    },
    "fieldName": "date / time",
    "fieldType": "date",
    "conceptID": "",
}



"""Dataclass to organize and convert the data type of a Record's detail."""

from dataclasses import dataclass
from typing import Any, Optional


@dataclass
class FieldType:
    """Organize and convert the data types of a Record's detail."""

    dropdown = "enum"
    numeric = "float"
    single_line = "freetext"
    multi_line = "blocktext"
    date_time = "date"
    geospatial = "geo"
    file_or_media_url = "file"
    record_pointer = "resource"
    relationship_marker = "relmarker"

    @classmethod
    def to_sql(cls, datatype: str) -> str:
        """
        Convert a Heurist data type label (i.e. "enum") to an SQL equivalent.

        Args:
            datatype (str): Heurist data type.

        Returns:
            str: SQL data type.
        """

        if datatype == cls.numeric:
            return "FLOAT"
        elif datatype == cls.date_time:
            return "DATE[2]"
        elif datatype == cls.record_pointer or datatype == cls.relationship_marker:
            return "INTEGER"
        elif datatype == cls.dropdown:
            return "VARCHAR"
        else:
            return "TEXT"

    @classmethod
    def from_detail(cls, detail: dict) -> str:
        """Extract the field type from a record's detail.

        Args:
            detail (dict): Record's detail.

        Returns:
            str: Field type name.
        """

        return detail["fieldType"]

    @classmethod
    def to_pydantic(cls, datatype: str) -> Any:
        """Convert Heurist field type to Python type.

        Args:
            datatype (str): Field type name.

        Returns:
            Any: Python type.
        """

        if datatype == cls.dropdown:
            return Optional[str]

        elif datatype == cls.numeric:
            return Optional[float]

        elif datatype == cls.single_line:
            return Optional[str]

        elif datatype == cls.multi_line:
            return Optional[str]

        elif datatype == cls.date_time:
            return dict

        elif datatype == cls.geospatial:
            return Optional[str]

        elif datatype == cls.file_or_media_url:
            return Optional[str]

        elif datatype == cls.record_pointer:
            return Optional[int]

        elif datatype == cls.relationship_marker:
            return Optional[str]

        else:
            return Optional[str]



import duckdb
from heurist.api.connection import HeuristAPIConnection
from heurist.database import TransformedDatabase
from heurist.utils.constants import DEFAULT_RECORD_GROUPS
from rich.progress import (
    BarColumn,
    MofNCompleteColumn,
    Progress,
    SpinnerColumn,
    TextColumn,
    TimeElapsedColumn,
)


def extract_transform_load(
    client: HeuristAPIConnection,
    duckdb_connection: duckdb.DuckDBPyConnection,
    user: tuple = (),
    record_group_names: tuple = DEFAULT_RECORD_GROUPS,
) -> None:
    """
    Workflow for (1) extracting, transforming, and loading the Heurist database \
        architecture into a DuckDB database and (2) extracting, transforming, \
        and loading record types' records into the created DuckDB database.

    Args:
        client (HeuristAPIConnection): Context of a Heurist API connection.
        duckdb_connection (duckdb.DuckDBPyConnection): Connection to a DuckDB database.
        user (tuple): IDs (integers) of targeted users.
        record_group_names (tuple): Names of the record group types. Must include at \
            least 1. Defaults to ("My record types").

    Returns:
        duckdb.DuckDBPyConnection: Open connection to the created DuckDB database.
    """

    # Export the Heurist database's structure
    with Progress(
        TextColumn("{task.description}"), SpinnerColumn(), TimeElapsedColumn()
    ) as p:
        _ = p.add_task("Get DB Structure")
        xml = client.get_structure()

    # Export individual record sets and insert into the DuckDB database
    with (
        Progress(
            TextColumn("{task.description}"),
            BarColumn(),
            MofNCompleteColumn(),
            TimeElapsedColumn(),
        ) as p,
    ):
        database = TransformedDatabase(
            conn=duckdb_connection,
            hml_xml=xml,
            record_type_groups=record_group_names,
        )
        t = p.add_task(
            "Get Records",
            total=len(database.pydantic_models.keys()),
        )
        for record_type in database.pydantic_models.values():
            rty_ID = record_type.rty_ID
            records = client.get_records(rty_ID, users=user)
            p.advance(t)
            database.insert_records(record_type_id=rty_ID, records=records)



from heurist.models.structural.DetailTypes import DetailTypes
from heurist.models.structural.dty import DTY
from heurist.models.structural.hml_structure import HMLStructure
from heurist.models.structural.RecStructure import RecStructure
from heurist.models.structural.RecTypeGroups import RecTypeGroups
from heurist.models.structural.RecTypes import RecTypes
from heurist.models.structural.rst import RST
from heurist.models.structural.rtg import RTG
from heurist.models.structural.rty import RTY

DetailTypes
DTY
HMLStructure
RecStructure
RecTypeGroups
RecTypes
RST
RTG
RTY



from heurist.models.structural.trm import TRM
from pydantic_xml import BaseXmlModel, element


class Terms(BaseXmlModel):
    """Dataclass for modeling all of the database structure's Terms.

    Attributes:
        trm (list): list of instantiated dataclasses that model all of the database's
            Terms.

    Examples:
        >>> from mock_data import DB_STRUCTURE_XML
        >>> from heurist.models.structural import HMLStructure
        >>>
        >>>
        >>> # Parse structure
        >>> xml = DB_STRUCTURE_XML
        >>> hml = HMLStructure.from_xml(xml)
        >>>
        >>> # Test class
        >>> first_detail_type = hml.Terms.trm[0]
        >>> first_detail_type.trm_ID
        12
    """

    trm: list[TRM] = element()



import unittest

from heurist.validators.record_validator import VALIDATION_LOG

from heurist.database.database import TransformedDatabase
from mock_data import DB_STRUCTURE_XML, RECORD_JSON


class DatabaseTest(unittest.TestCase):
    def setUp(self) -> None:
        self.db = TransformedDatabase(DB_STRUCTURE_XML)
        self.rectype = 103  # Story
        self.extracted_records = RECORD_JSON["heurist"]["records"]

    def tearDown(self):
        VALIDATION_LOG.unlink(missing_ok=True)

    def test(self):
        # Load the extracted records into the database
        rel = self.db.insert_records(
            records=self.extracted_records,
            record_type_id=self.rectype,
        )

        # Fetch the loaded records
        loaded_records = rel.fetchall()

        # Convert 1 of the records to a dictionary
        d = {k: v for k, v in zip(rel.columns, loaded_records[0])}

        # Confirm that the converted record's Heurist ID (H-ID) is an integer
        self.assertIsInstance(d["H-ID"], int)


if __name__ == "__main__":
    unittest.main()



from datetime import datetime
from typing import Literal, Optional

from pydantic_xml import BaseXmlModel, element


class TRM(BaseXmlModel, tag="trm", search_mode="unordered"):
    """Dataclass to model one of the database's vocabulary terms.

    When possible, the attribute descriptions are taken from Heurist's source code.
    """

    trm_ID: int = element()
    trm_Label: str = element()
    trm_InverseTermId: Optional[int] = element(default=None)
    trm_Description: Optional[str] = element(default=None)
    trm_Status: Literal["open", "approved", "reserved"] = element()
    trm_OriginatingDBID: int = element()
    trm_NameInOriginatingDB: Optional[str] = element(default=None)
    trm_IDInOriginatingDB: Optional[int] = element(default=None)
    trm_AddedByImport: bool = element()
    trm_IsLocalExtension: bool = element()
    trm_Domain: Literal["relation", "enum"] = element()
    trm_OntID: int = element()
    trm_ChildCount: int = element()
    trm_ParentTermID: Optional[int] = element(default=None)
    trm_Depth: int = element()
    trm_Modified: datetime = element()
    trm_LocallyModified: bool = element()
    trm_Code: Optional[str] = element(default=None)
    trm_SemanticReferenceURL: Optional[str] = element(default=None)
    trm_IllustrationURL: Optional[str] = element(default=None)
    trm_VocabularyGroupID: Optional[int] = element(default=None)
    trm_OrderInBranch: Optional[int] = element(default=None)



import unittest

import duckdb
from heurist.validators.record_validator import VALIDATION_LOG

from heurist.database.database import TransformedDatabase
from mock_data import DB_STRUCTURE_XML


class ModelingTest(unittest.TestCase):
    def setUp(self) -> None:
        self.db = TransformedDatabase(
            DB_STRUCTURE_XML,
            conn=duckdb.connect(),
        )
        group_id = (
            self.db.conn.table("rtg")
            .filter("rtg_Name like 'My record types'")
            .select("rtg_ID")
            .fetchone()[0]
        )
        self.record_types = [
            t[0]
            for t in self.db.conn.table("rty")
            .filter(f"rty_RecTypeGroupID = {group_id}")
            .select("rty_ID")
            .fetchall()
        ]

    def tearDown(self):
        VALIDATION_LOG.unlink(missing_ok=True)

    def test_all_my_records(self):
        for id in self.record_types:
            r = self.db.describe_record_schema(id)
            self.assertIsNotNone(r)


if __name__ == "__main__":
    unittest.main()



"""
CLI command for downloading one requested record type's data.
"""

import json
from pathlib import Path

from heurist.api.connection import HeuristAPIConnection
from heurist.api.credentials import CredentialHandler
from rich.progress import (
    Progress,
    SpinnerColumn,
    TextColumn,
    TimeElapsedColumn,
)


def rty_command(
    credentials: CredentialHandler,
    rty: int,
    outfile: Path | str | None,
):
    with (
        Progress(
            TextColumn("{task.description}"),
            SpinnerColumn(),
            TimeElapsedColumn(),
        ) as p,
        HeuristAPIConnection(
            db=credentials.get_database(),
            login=credentials.get_login(),
            password=credentials.get_password(),
        ) as client,
    ):
        _ = p.add_task(f"Get Records of type {rty}", total=1)
        records = client.get_records(rty)
    # If no records of this type have been entered, stop.
    if len(records) == 0:
        print("No records of this type have been entered.\nExiting program...")
        exit()

    # Else, write the records to a JSON file.
    if not outfile:
        outfile = f"RTY_{rty}.json"
    if not isinstance(outfile, Path):
        outfile = Path(outfile)
    print(f"Writing results to: {outfile}")
    with open(outfile, "w") as f:
        json.dump(records, f, indent=4)



# Result of joining the database schema tables
METADATA = {
    "dty_ID": 1111,
    "rst_DisplayName": "date_of_creation",
    "dty_Type": "date",
    "rst_MaxValues": 0,
}

# Result of a record's JSON export
DETAIL = [
    {
        "dty_ID": 1111,
        "value": {
            "start": {
                "earliest": "1180",
                "latest": "1231",
                "profile": "1",  # central
            },
            "end": {
                "latest": "1250",
                "earliest": "1246",
                "profile": "3",  # slowFinish
            },
            "determination": "2",  # conjecture
            "estMinDate": 1180,
            "estMaxDate": 1250.1231,
        },
        "fieldName": "date_of_creation",
        "fieldType": "date",
        "conceptID": "",
    },
    {
        "dty_ID": 1111,
        "value": {
            "start": {"earliest": "1454"},
            "end": {"latest": "1456"},
            "estMinDate": 1454,
            "estMaxDate": 1456.1231,
        },
        "fieldName": "date_of_creation",
        "fieldType": "date",
        "conceptID": "",
    },
]



from collections import Counter

from heurist.log import yield_log_blocks
from rich.table import Table


def log_summary(lines: list[str]) -> tuple[Table, Table]:
    rectypes = []
    recs = []
    for block in yield_log_blocks(lines):
        rectypes.append(block.recType)
        recs.append(block.recID)

    rectype_counter = Counter(rectypes)
    rec_counter = Counter(recs)

    rec_table = Table(
        title="Most problematic records",
        caption="Note: Invalid records are not saved in the DuckDB database.",
    )
    rec_table.add_column("Record ID", style="red")
    rec_table.add_column("Number of problems")
    for rec, count in rec_counter.most_common(10):
        rec_table.add_row(str(rec), str(count))

    type_table = Table(
        title="Types of invalid records",
        caption="Note: Invalid records are not saved in the DuckDB database.",
    )
    type_table.add_column("Record Type", style="red")
    type_table.add_column("Number of records")
    for rec, count in rectype_counter.items():
        type_table.add_row(str(rec), str(count))

    return type_table, rec_table



import unittest

from heurist.models.dynamic import HeuristRecord
from heurist.validators.record_validator import VALIDATION_LOG, RecordValidator

# Result of joining the database schema tables for a detail that takes only 1 value.
LIMITED_METADATA = [
    {
        "dty_ID": 1090,
        "rst_DisplayName": "language",
        "dty_Type": "enum",
        "rst_MaxValues": 1,
    }
]


# Result of joining the database schema tables for a detail that takes => 0 values.
REPEATED_METADATA = [
    {
        "dty_ID": 1090,
        "rst_DisplayName": "language",
        "dty_Type": "enum",
        "rst_MaxValues": 0,
    }
]

# Result of a record's JSON export
RECORDS = [
    {
        "rec_ID": 1001,
        "rec_RecTypeID": 100,
        "details": [
            # Repeated enum detail
            {
                "dty_ID": 1090,
                "value": "9728",
                "termLabel": "dum (Middle Dutch)",
                "termCode": "dum",
                "fieldName": "language",
                "fieldType": "enum",
                "conceptID": "",
            },
            {
                "dty_ID": 1090,
                "value": "9470",
                "termLabel": "fro (Old French)",
                "termCode": "fro",
                "fieldName": "language",
                "fieldType": "enum",
                "conceptID": "",
            },
        ],
    }
]


class ValidRepeatedDetailTest(unittest.TestCase):
    def test(self):
        hr = HeuristRecord(
            rty_ID=100, rty_Name="Example", detail_metadata=REPEATED_METADATA
        )
        validator = RecordValidator(
            rty_ID=100, pydantic_model=hr.model, records=RECORDS
        )
        # Assert that this process does NOT raise a warning that is logged to stdout.
        with self.assertNoLogs():
            for model in validator:
                break

        actual_record_type = model.type
        expected_record_type = 100
        self.assertEqual(actual_record_type, expected_record_type)

        expected_foreign_keys = [9728, 9470]
        actual_foreign_keys = model.DTY1090_TRM
        self.assertListEqual(expected_foreign_keys, actual_foreign_keys)

    def tearDown(self):
        if VALIDATION_LOG.is_file():
            VALIDATION_LOG.unlink()
        return super().tearDown()


class InvalidRepeatedDetaiTest(unittest.TestCase):
    def test(self):
        """Run the validation process on records that should log an error."""
        hr = HeuristRecord(
            rty_ID=100, rty_Name="Example", detail_metadata=LIMITED_METADATA
        )
        validator = RecordValidator(
            rty_ID=100, pydantic_model=hr.model, records=RECORDS
        )
        # Assert that this validation raises a warning that is logged to stdout.
        with self.assertLogs():
            for model in validator:
                actual_foreign_key = model.DTY1090_TRM
                self.assertIsNone(actual_foreign_key)

    def tearDown(self):
        if VALIDATION_LOG.is_file():
            VALIDATION_LOG.unlink()
        return super().tearDown()


if __name__ == "__main__":
    unittest.main()



import duckdb
from heurist.schema.models import DTY, RTY
from heurist.utils.rel_to_dict_array import rel_to_dict_array


def convert_rty_description(description: duckdb.DuckDBPyRelation) -> dict:
    """
    Convert the SQL result of the joined record type schema description into \
    a Python dictionary.

    Args:
        description (duckdb.DuckDBPyRelation): Relation from SQL query.

    Returns:
        dict: Dictionary representation of a record's schema.
    """

    rel = description.filter("dty_Type not like 'separator'").order(
        "group_id asc, rst_DisplayOrder asc"
    )

    sections = {}

    for field in rel_to_dict_array(rel):
        section_id = field["group_id"]
        section_name = field["sec"]
        if not sections.get(section_id):
            sections.update({section_id: {"sectionName": section_name, "fields": []}})
        field_model = DTY.model_validate(field).model_dump()
        sections[section_id]["fields"].append(field_model)

    section_list = list(sections.values())

    rty_rel = rel.limit(1)
    rty_data = rel_to_dict_array(rty_rel)[0]
    rty = RTY.model_validate(rty_data)

    output = {rty.rty_ID: {"metadata": rty.model_dump(), "sections": section_list}}
    return output



# Result of joining the database schema tables
METADATA = {
    "dty_ID": 1111,
    "rst_DisplayName": "date_of_creation",
    "dty_Type": "date",
    "rst_MaxValues": 1,
}

# Result of a record's JSON export
DETAIL = {
    "dty_ID": 1111,
    "value": "2024-03-19",
    "fieldName": "date_of_creation",
    "fieldType": "date",
    "conceptID": "",
}



"""Class to compose URIs for calling the Heurist API."""

from typing import Literal

from heurist.api.constants import (
    HUMA_NUM_SERVER,
    RECORD_JSON_EXPORT_PATH,
    RECORD_XML_EXPORT_PATH,
    STRUCTURE_EXPORT_PATH,
)

COMMA = "%2C"
COLON = "%3A"


class URLBuilder:
    """Class to construct endpoints for the Heurist API (on Huma-Num's server)."""

    def __init__(self, database_name: str, server: str = HUMA_NUM_SERVER) -> None:
        self.server = server
        self.database_name = database_name

    @property
    def db_api(self) -> str:
        return f"{self.server}{STRUCTURE_EXPORT_PATH}"

    @property
    def xml_record_api(self) -> str:
        return f"{self.server}{RECORD_XML_EXPORT_PATH}"

    @property
    def json_record_api(self) -> str:
        return f"{self.server}{RECORD_JSON_EXPORT_PATH}"

    @classmethod
    def _join_queries(cls, *args) -> str:
        """Join 1 or more queries together with an ampersand.

        Returns:
            str: Fragment of a path for the URL.
        """
        return "&".join([a for a in args if a is not None])

    @classmethod
    def _join_list_items(cls, *args) -> str:
        """Join 1 or more items in a list of queries.

        Examples:
            >>> item1 = '{"filter"%3A"value"}'
            >>> item2 = '{"filter"%3A"value"}'
            >>> item3 = '{"filter"%3A"value"}'
            >>> URLBuilder._join_list_items(item1, item2, item3)
            '[{"filter"%3A"value"}%2C{"filter"%3A"value"}%2C{"filter"%3A"value"}]'

        Returns:
            str: Fragment of a path for the URL.
        """
        start = "["
        end = "]"
        items = COMMA.join([a for a in args if a is not None])
        return f"{start}{items}{end}"

    @classmethod
    def _make_filter_obj(cls, filter: str, value: str | int) -> str:
        start = "{"
        end = "}"
        return f'{start}"{filter}"{COLON}"{value}"{end}'

    @classmethod
    def _join_comma_separated_values(cls, *args) -> str:
        return COMMA.join([str(a) for a in args if a is not None])

    def get_db_structure(self) -> str:
        """
        URL to retrieve the database structure.

        Examples:
            >>> db = "mock_db"
            >>> builder = URLBuilder(db)
            >>> builder.get_db_structure()
            'https://heurist.huma-num.fr/heurist/hserv/structure/export/getDBStructureAsXML.php?db=mock_db'

        Returns:
            str: URL to retrieve the database structure.
        """
        db = f"?db={self.database_name}"
        return f"{self.db_api}{db}"

    def get_records(
        self,
        record_type_id: int,
        form: Literal["xml", "json"] = "xml",
        users: tuple = (),
    ) -> str:
        """Build a URL to retrieve records of a certain type.

        Examples:
            >>> db = "mock_db"
            >>> builder = URLBuilder(db)
            >>> builder.get_records(101)
            'https://heurist.huma-num.fr/heurist/export/xml/flathml.php?q=[{"t"%3A"101"}%2C{"sortby"%3A"t"}]&a=1&db=mock_db&depth=all&linkmode=direct_links'

            >>> db = "mock_db"
            >>> builder = URLBuilder(db)
            >>> builder.get_records(102, form="json")
            'https://heurist.huma-num.fr/heurist/hserv/controller/record_output.php?q=[{"t"%3A"102"}%2C{"sortby"%3A"t"}]&a=1&db=mock_db&depth=all&linkmode=direct_links&format=json&defs=0&extended=2'

            >>> db = "mock_db"
            >>> builder = URLBuilder(db)
            >>> builder.get_records(102, users=(2,16,))
            'https://heurist.huma-num.fr/heurist/export/xml/flathml.php?q=[{"t"%3A"102"}%2C{"sortby"%3A"t"}%2C{"addedby"%3A"2%2C16"}]&a=1&db=mock_db&depth=all&linkmode=direct_links'


        Args:
            record_type_id (int): Heurist ID of the record type.
            form (Literal["xml", "json"]): The format of the exported data.

        Returns:
            str: URL to retrieve records of a certain type.
        """

        a = "a=1"
        db = "db=%s" % (self.database_name)
        depth = "depth=all"
        link_mode = "linkmode=direct_links"

        if form == "json":
            api = self.json_record_api
            format_args = "format=json&defs=0&extended=2"
        else:
            api = self.xml_record_api
            format_args = None

        # Make the query based on parameters
        record_type_filter = self._make_filter_obj(filter="t", value=record_type_id)
        sortby_filter = self._make_filter_obj(filter="sortby", value="t")
        if len(users) > 0:
            user_string = self._join_comma_separated_values(*users)
            users_filter = self._make_filter_obj(filter="addedby", value=user_string)
        else:
            users_filter = None
        query_path = self._join_list_items(
            record_type_filter, sortby_filter, users_filter
        )
        query = f"?q={query_path}"

        path = self._join_queries(query, a, db, depth, link_mode, format_args)
        return f"{api}{path}"



import duckdb


def rel_to_dict_array(rel: duckdb.DuckDBPyRelation) -> list[dict]:
    output = []
    for row in rel.fetchall():
        output.append({k: v for k, v in zip(rel.columns, row)})
    return output



import unittest
from datetime import datetime

from heurist.validators.detail_validator import DetailValidator

URI = "https://upload.wikimedia.org/wikipedia/commons/9/9a/Gull_portrait_ca_usa.jpg"


DATE_DEFAULT = {
    "comment": None,
    "value": None,
    "start": {
        "earliest": None,
        "latest": None,
        "estProfile": None,
        "estDetermination": None,
    },
    "end": {
        "earliest": None,
        "latest": None,
        "estProfile": None,
        "estDetermination": None,
    },
    "estDetermination": None,
    "estProfile": None,
    "timestamp": {"in": None, "type": None, "circa": False},
    "estMinDate": None,
    "estMaxDate": None,
}


class TestFile(unittest.TestCase):
    def setUp(self):
        from mock_data.file.single import DETAIL

        self.detail = DETAIL

    def test(self):
        expected = URI
        actual = DetailValidator.validate_file(self.detail)
        self.assertEqual(expected, actual)


class TestEnum(unittest.TestCase):
    def setUp(self):
        from mock_data.enum.single import DETAIL

        self.detail = DETAIL

    def test(self):
        expected = "dum (Middle Dutch)"
        actual = DetailValidator.validate_enum(self.detail)
        self.assertEqual(expected, actual)


class TestGeo(unittest.TestCase):
    def setUp(self):
        from mock_data.geo.single import DETAIL_POINT, DETAIL_POLYGON

        self.point_detail = DETAIL_POINT
        self.polygon_detail = DETAIL_POLYGON

    def test_point(self):
        expected = "POINT(2.19726563 48.57478991)"
        actual = DetailValidator.validate_geo(self.point_detail)
        self.assertEqual(expected, actual)

    def test_polygon(self):
        startswith = "POLYGON((-2.82548747 55.12653961,-2.82912354 55.12473717,\
            -2.83104469 55.12336704,"
        expected = startswith[:20]
        actual = DetailValidator.validate_geo(self.polygon_detail)[:20]
        self.assertEqual(expected, actual)


class TestDate(unittest.TestCase):
    maxDiff = None

    def setUp(self):
        self.expected_result = DATE_DEFAULT.copy()
        return super().setUp()

    def test_simple_date(self):
        detail = {
            "dty_ID": 1000,
            "value": "2024-03-19",
        }
        self.expected_result.update(
            {
                "value": datetime(2024, 3, 19, 0, 0),
            }
        )
        actual_result = DetailValidator.validate_date(detail)
        self.assertDictEqual(actual_result, self.expected_result)

    def test_fuzzy_date(self):
        detail = {
            "value": {
                "start": {
                    "earliest": "1180",
                    "latest": "1231",
                    "profile": "1",
                },
                "end": {
                    "latest": "1250",
                    "earliest": "1246",
                    "profile": "3",
                },
                "determination": "2",
                "estMinDate": 1180,
                "estMaxDate": 1250.1231,
            },
        }
        self.expected_result.update(
            {
                "start": {
                    "earliest": datetime(1180, 1, 1, 0, 0),
                    "latest": datetime(1231, 1, 1, 0, 0),
                    "estProfile": "central",
                    "estDetermination": None,  # Don't forget to keep this null
                },
                "end": {
                    "earliest": datetime(1246, 1, 1, 0, 0),
                    "latest": datetime(1250, 1, 1, 0, 0),
                    "estProfile": "slowFinish",
                    "estDetermination": None,  # Don't forget to keep this null
                },
                "estDetermination": "conjecture",
                "estMinDate": datetime(1180, 1, 1, 0, 0),
                "estMaxDate": datetime(1250, 12, 31, 0, 0),
            }
        )
        actual_result = DetailValidator.validate_date(detail)
        self.assertDictEqual(actual_result, self.expected_result)


class TestResource(unittest.TestCase):
    def setUp(self):
        from mock_data.resource.single import DETAIL

        self.detail = DETAIL

    def test(self):
        expected = 36
        actual = DetailValidator.validate_resource(self.detail)
        self.assertEqual(expected, actual)


if __name__ == "__main__":
    unittest.main()



import csv
import json
import unittest
from pathlib import Path

from heurist.api.credentials import CredentialHandler
from heurist.cli.schema import schema_command
from heurist.validators.record_validator import VALIDATION_LOG


class OnlineSchemaCommand(unittest.TestCase):
    tempdir = Path(__file__).parent.joinpath("temp")
    tempfile_json = tempdir.joinpath("recordTypes.json")

    def setUp(self):
        self.tempdir.mkdir(exist_ok=True)
        try:
            self.credentials = CredentialHandler()
        except SystemExit:
            self.skipTest(
                "Connection could not be established.\nCannot test client without \
                    database connection."
            )

    def tearDown(self):
        for f in self.tempdir.iterdir():
            f.unlink(missing_ok=True)
        self.tempdir.rmdir()
        VALIDATION_LOG.unlink(missing_ok=True)
        return super().tearDown()

    def json(self):
        _ = schema_command(
            credentials=self.credentials,
            record_group=["My record types"],
            outdir=self.tempdir,
            output_type="json",
            debugging=True,
        )
        with open(self.tempfile_json) as f:
            data = json.load(f)
        actual = len(data["items"])
        self.assertGreater(actual, 0)

    def csv(self):
        _ = schema_command(
            credentials=self.credentials,
            record_group=["My record types"],
            outdir=self.tempdir,
            output_type="csv",
            debugging=True,
        )
        for file in self.tempdir.iterdir():
            with open(file, mode="r") as f:
                reader = csv.DictReader(f)
                row_count = len([_ for r in reader])
                self.assertGreater(row_count, 0)


if __name__ == "__main__":
    unittest.main()



"""Generate the code reference pages."""

from pathlib import Path

import mkdocs_gen_files

nav = mkdocs_gen_files.Nav()

root = Path(__file__).parent.parent
package = root.joinpath("heurist")

# Recursively get python files in the package directory
for path in sorted(package.rglob("*.py")):
    # Parse the python module's path
    module_path = path.relative_to(package).with_suffix("")
    # Build the
    doc_path = path.relative_to(package).with_suffix(".md")
    full_doc_path = Path("reference", doc_path)

    parts = tuple(module_path.parts)

    if parts[0] == "mock_data" or parts[0] == "cli":
        continue
    if parts[-1] == "constants.py":
        continue

    if parts[-1] == "__init__":
        parts = parts[:-1]
    elif parts[-1] == "__main__":
        continue
    elif parts[0] == "__init__":
        continue
    if str(full_doc_path).endswith("__init__.md"):
        continue

    nav[parts] = doc_path.as_posix()

    with mkdocs_gen_files.open(full_doc_path, "w") as fd:
        identifier = ".".join(parts)
        print("::: " + identifier, file=fd)

    mkdocs_gen_files.set_edit_path(full_doc_path, path.relative_to(root))

    with mkdocs_gen_files.open("reference/SUMMARY.md", "w") as nav_file:
        nav_file.writelines(nav.build_literate_nav())



# Result of joining the database schema tables
METADATA = {
    "dty_ID": 1109,
    "rst_DisplayName": "preferred_name",
    "dty_Type": "freetext",
    "rst_MaxValues": 1,
}

# Result of a record's JSON export
DETAIL = {
    "dty_ID": 1109,
    "value": "name of something",
    "fieldName": "single line",
    "fieldType": "freetext",
    "conceptID": "",
}



import os

from dotenv import find_dotenv, load_dotenv
from heurist.api.connection import HeuristAPIConnection
from heurist.api.exceptions import MissingParameterException


class CredentialHandler:
    env_file = find_dotenv()
    db_key = "DB_NAME"
    login_key = "DB_LOGIN"
    password_key = "DB_PASSWORD"

    def __init__(
        self,
        database_name: str | None = None,
        login: str | None = None,
        password: str | None = None,
        debugging: bool = False,
    ):
        if not debugging:
            load_dotenv(self.env_file)

        params = [
            (self.db_key, database_name),
            (self.login_key, login),
            (self.password_key, password),
        ]

        # Set all the secret variables in the environment
        for key, var in params:
            if var:
                self.set_var(key=key, var=var)
            # Confirm that the environment variable is set
            v = self.get_var(key=key)
            if not v or v == "":
                e = MissingParameterException(parameter=key, env_file=self.env_file)
                raise SystemExit(e)

    def test_connection(self) -> None:
        with HeuristAPIConnection(
            db=self.get_database(), login=self.get_login(), password=self.get_password()
        ) as _:
            pass

    @classmethod
    def _reset_envvars(cls) -> None:
        keys = [cls.db_key, cls.login_key, cls.password_key]
        for key in keys:
            if os.environ.get(key):
                os.environ.pop(key)

    @classmethod
    def set_var(cls, key: str, var: str) -> None:
        os.environ[key] = var

    @classmethod
    def get_var(cls, key: str) -> str | KeyError:
        return os.getenv(key)

    @classmethod
    def get_database(cls) -> str:
        return cls.get_var(key="DB_NAME")

    @classmethod
    def get_login(cls) -> str:
        return cls.get_var(key="DB_LOGIN")

    @classmethod
    def get_password(cls) -> str:
        return cls.get_var(key="DB_PASSWORD")



# Result of joining the database schema tables
METADATA = {
    "dty_ID": 1111,
    "rst_DisplayName": "date_of_creation",
    "dty_Type": "date",
    "rst_MaxValues": 0,
}

# Result of a record's JSON export
DETAIL = [
    {
        "dty_ID": 1111,
        "value": "2024-03-19",
        "fieldName": "date_of_creation",
        "fieldType": "date",
        "conceptID": "",
    },
    {
        "dty_ID": 1111,
        "value": {
            "timestamp": {"in": "1188", "type": "s", "circa": True},
            "comment": "1188",
            "estMinDate": 1188,
            "estMaxDate": 1188,
        },
        "fieldName": "date_of_creation",
        "fieldType": "date",
        "conceptID": "",
    },
]



from heurist.models.structural.dty import DTY
from pydantic_xml import BaseXmlModel, element


class DetailTypes(BaseXmlModel):
    """Dataclass for modeling all of the database structure's Detail Types.

    Attributes:
        dty (list): list of instantiated dataclasses that model all of the database's
            Detail Types.

    Examples:
        >>> from mock_data import DB_STRUCTURE_XML
        >>> from heurist.models.structural import HMLStructure
        >>>
        >>>
        >>> # Parse structure
        >>> xml = DB_STRUCTURE_XML
        >>> hml = HMLStructure.from_xml(xml)
        >>>
        >>> # Test class
        >>> first_detail_type = hml.DetailTypes.dty[0]
        >>> first_detail_type.dty_ID
        1
        >>> singular_pointer = [d for d in hml.DetailTypes.dty if d.dty_ID == 1295][0]
        >>> singular_pointer.dty_PtrTargetRectypeIDs
        [101]
        >>> plural_pointer = [d for d in hml.DetailTypes.dty if d.dty_ID == 1256][0]
        >>> plural_pointer.dty_PtrTargetRectypeIDs
        [101, 105, 106]

    """

    dty: list[DTY] = element()



from datetime import datetime
from typing import Annotated, Optional

from heurist.validators import parse_heurist_date
from pydantic import BaseModel, BeforeValidator, Field

PROFILE_MAP = {"0": "flat", "1": "central", "2": "slowStart", "3": "slowFinish"}
DETERMINATION_MAP = {
    "0": "unknown",
    "1": "attested",
    "2": "conjecture",
    "3": "measurement",
}


def parse_profile(value) -> str | None:
    if PROFILE_MAP.get(value):
        return PROFILE_MAP[value]


def parse_determination(value) -> str | None:
    if DETERMINATION_MAP.get(value):
        return DETERMINATION_MAP[value]


HeuristDate = Annotated[datetime, BeforeValidator(parse_heurist_date)]
HeuristProfile = Annotated[str, BeforeValidator(parse_profile)]
HeuristDetermination = Annotated[str, BeforeValidator(parse_determination)]


class DateLimit(BaseModel):
    earliest: Optional[HeuristDate] = Field(default=None)
    latest: Optional[HeuristDate] = Field(default=None)
    estProfile: Optional[HeuristProfile] = Field(
        default=None, validation_alias="profile"
    )
    estDetermination: Optional[HeuristDetermination] = Field(
        default=None, validation_alias="determination"
    )


class Timestamp(BaseModel):
    inYear: Optional[HeuristDate] = Field(default=None, alias="in")
    typeTime: Optional[str] = Field(default=None, alias="type")
    circa: Optional[bool] = Field(default=False)


class TemporalObject(BaseModel):
    comment: Optional[str] = Field(default=None)
    value: Optional[HeuristDate] = Field(default=None)
    start: Optional[DateLimit] = Field(default=DateLimit(**{}))
    end: Optional[DateLimit] = Field(default=DateLimit(**{}))
    estDetermination: Optional[HeuristDetermination] = Field(
        default=None, validation_alias="determination"
    )
    estProfile: Optional[HeuristProfile] = Field(
        default=None, validation_alias="profile"
    )
    timestamp: Optional[Timestamp] = Field(default=Timestamp(**{}))
    estMinDate: Optional[HeuristDate] = Field(default=None)
    estMaxDate: Optional[HeuristDate] = Field(default=None)



# Result of joining the database schema tables
METADATA = {
    "dty_ID": 1090,
    "rst_DisplayName": "language",
    "dty_Type": "enum",
    "rst_MaxValues": 1,
}

# Result of a record's JSON export
DETAIL = {
    "dty_ID": 1090,
    "value": "9728",
    "termLabel": "dum (Middle Dutch)",
    "termCode": "dum",
    "fieldName": "language",
    "fieldType": "enum",
    "conceptID": "",
}



"""Module for utilities commonly used by XML parsers in schemas."""

import json


def split_ids(input: str) -> list:
    """Function for converting a string representation of a list of quoted integers \
        into a Python list object.

    Examples:
        >>> s = "3001,3110,3113,3288"
        >>> split_ids(s)
        [3001, 3110, 3113, 3288]
        >>> s = '[\\"3001\\",\\"3110\\",\\"3113\\",\\"3288\\"]'
        >>> split_ids(s)
        [3001, 3110, 3113, 3288]
        >>> l = ['[\\"3001\\",\\"3110\\",\\"3113\\",\\"3288\\"]']
        >>> split_ids(l)
        [3001, 3110, 3113, 3288]
        >>> s = '[]'
        >>> split_ids(s)
        []

    Args:
        input (str|Any): List of integers.

    Returns:
        list: _description_
    """

    # Handle input of potentially malformatted JSON list
    if isinstance(input, list) or "[" in input:
        if not isinstance(input, str):
            s = json.dumps(input)
            if not isinstance(s, str):
                print(input)
                raise TypeError("Integer list was not converted to string.")
        else:
            s = input

        # Remove the JSON string escape characters
        s = s.replace("\\", "")
        s = s.replace('"', "")
        s = s.replace("[", "")
        s = s.replace("]", "")

        input_string = s

    else:
        input_string = input

    # Split and recast regular string
    return [int(i) for i in input_string.split(",") if i != ""]



"""
Test of the API client.
Requires a connection to a remote Heurist database.

Declare the login credentials in a .env file
"""

import unittest

from heurist.api.connection import HeuristAPIConnection
from heurist.api.credentials import CredentialHandler
from heurist.api.exceptions import AuthenticationError
from lxml import etree
from requests.exceptions import ConnectTimeout

TEST_RECORD_TYPE = 103
TEST_USER = 2


class CredentialHandlerTest(unittest.TestCase):
    def setUp(self) -> None:
        CredentialHandler._reset_envvars()
        try:
            self.credentials = CredentialHandler()
        except SystemExit:
            self.skipTest(
                "Connection could not be established.\nCannot test client without \
                    database connection."
            )

    def test_user_filter(self):
        """Test the API client's ability to extract records created by a \
            certain user."""

        with HeuristAPIConnection(
            db=self.credentials.get_database(),
            login=self.credentials.get_login(),
            password=self.credentials.get_password(),
        ) as client:
            try:
                records = client.get_records(
                    record_type_id=TEST_RECORD_TYPE, users=(TEST_USER,)
                )
            except AuthenticationError:
                self.skipTest(
                    "Connection could not be established.\nCannot test client without \
                        database connection."
                )

        # Confirm that every record was made by the targeted user
        for record in records:
            expected = str(TEST_USER)
            actual = record.get("rec_AddedByUGrpID")
            self.assertEqual(expected, actual)


class ClientUnitTest(unittest.TestCase):
    def setUp(self) -> None:
        CredentialHandler._reset_envvars()
        try:
            self.credentials = CredentialHandler()
        except SystemExit:
            self.skipTest(
                "Connection could not be established.\nCannot test client without \
                    database connection."
            )

        try:
            with HeuristAPIConnection(
                db=self.credentials.get_database(),
                login=self.credentials.get_login(),
                password=self.credentials.get_password(),
            ) as _:
                pass
        except ConnectTimeout:
            self.skipTest(
                "Connection could not be established.\nCannot test client without \
                        database connection."
            )

    def test_hml_export(self):
        """Test the API client's ability to extract the database schema."""

        with HeuristAPIConnection(
            db=self.credentials.get_database(),
            login=self.credentials.get_login(),
            password=self.credentials.get_password(),
        ) as client:
            # Confirm that the client receives bytes data.
            hml_bytes = client.get_structure()
            self.assertIsInstance(hml_bytes, bytes)

        # Confirm that the data is the <hml_structure> XML.
        root = etree.fromstring(hml_bytes)
        expected = "hml_structure"
        actual = root.tag
        self.assertEqual(expected, actual)

    def test_json_records(self):
        """Test the API client's ability to extract records in a JSON array."""
        with HeuristAPIConnection(
            db=self.credentials.get_database(),
            login=self.credentials.get_login(),
            password=self.credentials.get_password(),
        ) as client:
            records = client.get_records(record_type_id=TEST_RECORD_TYPE)

        # Confirm that the data is a JSON array.
        self.assertIsInstance(records, list)
        self.assertIsInstance(records[0], dict)

    def test_xml_records(self):
        """Test the API client's ability to extract records in XML bytes"""

        with HeuristAPIConnection(
            db=self.credentials.get_database(),
            login=self.credentials.get_login(),
            password=self.credentials.get_password(),
        ) as client:
            # Confirm that the client receives bytes data.
            records = client.get_records(record_type_id=TEST_RECORD_TYPE, form="xml")

        self.assertIsInstance(records, bytes)

        # Confirm that the data is a record XML.
        root = etree.fromstring(records)
        expected = r"{https://heuristnetwork.org}hml"
        actual = root.tag
        self.assertEqual(expected, actual)


if __name__ == "__main__":
    unittest.main()



from heurist.models.structural.rty import RTY
from pydantic_xml import BaseXmlModel, element


class RecTypes(BaseXmlModel):
    """Dataclass for modeling all of the database structure's Record Types.

    Attributes:
        rty (list): list of instantiated dataclasses that model all of the database's
            Record Types.

    Examples:
        >>> from mock_data import DB_STRUCTURE_XML
        >>> from heurist.models.structural import HMLStructure
        >>>
        >>>
        >>> # Parse structure
        >>> xml = DB_STRUCTURE_XML
        >>> hml = HMLStructure.from_xml(xml)
        >>>
        >>> # Test class
        >>> first_record_type = hml.RecTypes.rty[0]
        >>> first_record_type.rty_ID
        1
    """

    rty: list[RTY] = element()



from heurist.models.dynamic.annotation import PydanticField
from heurist.sql.sql_safety import SafeSQLName
from pydantic import BaseModel, Field
from pydantic import create_model as pydantic_create_model


class HeuristRecord:
    def __init__(self, rty_Name: str, rty_ID: int, detail_metadata: list[dict]):
        self.rty_ID = rty_ID
        # Create an SQL-safe name to give the model when it's serialized to a table
        self.table_name = SafeSQLName().create_table_name(record_name=rty_Name)
        # Create the Pydantic model
        self.model = create_record_type_model(
            model_name=self.table_name, detail_metadata=detail_metadata
        )


def create_record_type_model(model_name: str, detail_metadata: list[dict]) -> BaseModel:
    """
    Each detail's set of metadata must be a dictionary with the following keys:
        dty_ID,
        rst_DisplayName,
        dty_Type,
        rst_MaxValues

    Examples:
        >>> d1 = {'dty_ID': 1,
        ... 'rst_DisplayName': 'label',
        ... 'dty_Type': 'enum',
        ... 'rst_MaxValues': 1}
        >>> d2 = {'dty_ID': 2,
        ... 'rst_DisplayName': 'date',
        ... 'dty_Type': 'date',
        ... 'rst_MaxValues': 1}
        >>> model = create_record_type_model('test', [d1, d2])
        >>> model.__annotations__.keys()
        dict_keys(['id', 'type', 'DTY1', 'DTY1_TRM', 'DTY2'])

    Args:
        model_name (str): The model's name.
        detail_metadata (list[dict]): A list of metadata about each field (detail).

    Return:
        (BaseModel): Pydantic model.
    """

    # Indifferent to the record's data fields, set up universal data fields
    # present in every dynamic Pydantic model for records of any type
    kwargs = {
        "id": (
            int,
            Field(
                default=0,
                alias="rec_ID",
                validation_alias="rec_ID",
                serialization_alias="H-ID",
            ),
        ),
        "type": (
            int,
            Field(
                default=0,
                alias="rec_RecTypeID",
                validation_alias="rec_RecTypeID",
                serialization_alias="type_id",
            ),
        ),
    }

    # Convert each of the record's details into a Pydantic kwarg
    for detail in detail_metadata:
        # Add the field's default parsed value
        annotation = PydanticField(**detail)
        field = annotation.build_field()
        kwargs.update(field)

        if detail["dty_Type"] == "enum":
            field = annotation.build_term_fk()
            kwargs.update(field)

    # Using Pydantic's 'create_model' module, build the dynamic model
    return pydantic_create_model(model_name, **kwargs)



# Result of joining the database schema tables
METADATA = {
    "dty_ID": 1090,
    "rst_DisplayName": "language",
    "dty_Type": "enum",
    "rst_MaxValues": 0,
}

# Result of a record's JSON export
DETAIL = [
    {
        "dty_ID": 1090,
        "value": "9728",
        "termLabel": "dum (Middle Dutch)",
        "termCode": "dum",
        "fieldName": "language",
        "fieldType": "enum",
        "conceptID": "",
    },
    {
        "dty_ID": 1090,
        "value": "9470",
        "termLabel": "fro (Old French)",
        "termCode": "fro",
        "fieldName": "language",
        "fieldType": "enum",
        "conceptID": "",
    },
]



"""Constant text variables for Heurist API."""

import os

HUMA_NUM_SERVER = "https://heurist.huma-num.fr/heurist"

RECORD_XML_EXPORT_PATH = "/export/xml/flathml.php"

RECORD_JSON_EXPORT_PATH = "/hserv/controller/record_output.php"

STRUCTURE_EXPORT_PATH = "/hserv/structure/export/getDBStructureAsXML.php"

timeout_var = os.environ.get("READTIMEOUT", 10)
if isinstance(timeout_var, str):
    timeout_var = int(timeout_var)

READTIMEOUT = timeout_var

MAX_RETRY = 3



import os
import unittest

import pytest
from heurist.api.connection import HeuristAPIConnection
from heurist.api.credentials import CredentialHandler

IN_GITHUB_ACTIONS = os.getenv("GITHUB_ACTIONS") == "true"


@pytest.mark.skipif(
    IN_GITHUB_ACTIONS, reason="Do not run connection test in GitHub Actions."
)
class ConnectionWithEnvVars(unittest.TestCase):
    def setUp(self):
        """Reset environment variables"""
        CredentialHandler._reset_envvars()
        return super().setUp()

    def test_database_env_var(self):
        database = CredentialHandler(database_name=None).get_database()
        self.assertIsNotNone(database)

    def test_database_connection(self):
        credentials_from_env_vars = CredentialHandler()
        credentials_from_env_vars.test_connection()

    def test_client_request(self):
        credentials = CredentialHandler(database_name="api_dev")
        with HeuristAPIConnection(
            db=credentials.get_database(),
            login=credentials.get_login(),
            password=credentials.get_password(),
        ) as client:
            response = client.get_records(form="json", record_type_id=103)
        expected = "Symphony No. 1 in D major"
        actual = [i["rec_Title"] for i in response if i["rec_ID"] == "279"][0]
        self.assertEqual(expected, actual)

    def tearDown(self):
        """Reset environment variables"""
        CredentialHandler._reset_envvars()
        return super().tearDown()


class ConnectionWithoutEnvVars(unittest.TestCase):
    def setUp(self):
        """Reset environment variables"""
        CredentialHandler._reset_envvars()
        return super().setUp()

    def test_missing_login(self):
        with pytest.raises(SystemExit):
            CredentialHandler(login=None, debugging=True)

    def test_missing_db_name(self):
        with pytest.raises(SystemExit):
            CredentialHandler(database_name=None, debugging=True)

    def test_missing_password(self):
        with pytest.raises(SystemExit):
            CredentialHandler(password=None, debugging=True)

    def test_invalid_database_credentials(self):
        with pytest.raises(SystemExit):
            CredentialHandler(
                database_name="test_db",
                login="test_user",
                password="test_pass",
            ).test_connection()

    def tearDown(self):
        """Reset environment variables"""
        CredentialHandler._reset_envvars()
        return super().tearDown()


if __name__ == "__main__":
    unittest.main()



from datetime import datetime

import dateutil.parser
import dateutil.relativedelta


def parse_heurist_date(repr: str | int | float | None) -> datetime | None:
    """
    Convert Heurist's partial date representations to an ISO string format.

    Examples:
        >>> # Test a string representation of a date
        >>> v = "2024-03-19"
        >>> parse_heurist_date(v)
        datetime.datetime(2024, 3, 19, 0, 0)

        >>> # Test an integer representation of a year, i.e. circa 1188
        >>> v = 1188
        >>> parse_heurist_date(v)
        datetime.datetime(1188, 1, 1, 0, 0)

        >>> # Test a float representation of a date
        >>> v = 1250.1231
        >>> parse_heurist_date(v)
        datetime.datetime(1250, 12, 31, 0, 0)

    Args:
        repr (str | int | float): Heurist representation \
            of a date.

    Returns:
        datetime | None: Parsed date.
    """

    if not repr:
        return

    # Affirm Heurist's representation of the date is a Python string
    repr = str(repr)

    # If the Heurist representation is a year, change it to the start of
    # the year.
    if len(repr) <= 4:
        iso_str = f"{repr}-01-01"
        return dateutil.parser.parse(iso_str)

    # If the Heurist representation is a float, parse the month and day
    # shown after the decimal.
    elif "." in repr:
        splits = repr.split(".")
        year, smaller_than_year = splits[0], splits[1]
        if len(smaller_than_year) == 2:
            iso_str = f"{year}-{smaller_than_year}-01"
        elif len(smaller_than_year) == 4:
            iso_str = f"{year}-{smaller_than_year[:2]}-{smaller_than_year[2:]}"
        else:
            raise ValueError(repr)
        return dateutil.parser.parse(iso_str)

    # If the Heurist representation is a year and month, add the day
    # (first of the month)
    parts = repr.split("-")
    if len(parts) == 2:
        iso_str = f"{repr}-01"
        return dateutil.parser.parser(iso_str)

    # If no other conditions have been met, the representation is already in
    # ISO format YYYY-MM-DD.
    else:
        iso_str = repr
        return dateutil.parser.parse(iso_str)



from typing import Generator

from .model import LogDetail


def yield_log_blocks(lines: list[str]) -> Generator[LogDetail, None, None]:
    line_iterator = iter(lines)
    l1 = next(line_iterator, None)
    while l1 is not None:
        if l1 and not l1.startswith("\t"):
            l2 = next(line_iterator)
            l3 = next(line_iterator)
            l4 = next(line_iterator)
            l5 = next(line_iterator)
            yield LogDetail.load_lines(l1, l2, l3, l4, l5)
        l1 = next(line_iterator, None)



import duckdb
import polars as pl
from duckdb import DuckDBPyConnection, DuckDBPyRelation
from heurist.models.structural.hml_structure import HMLStructure
from heurist.sql import RECORD_TYPE_SCHEMA
from pydantic_xml import BaseXmlModel


class HeuristDatabase:
    """Base class for loading the original Heurist database structure."""

    BASE_TABLES = [
        ("rtg", "RecTypeGroups"),
        ("rst", "RecStructure"),
        ("rty", "RecTypes"),
        ("dty", "DetailTypes"),
        ("trm", "Terms"),
    ]

    def __init__(
        self,
        hml_xml: bytes,
        conn: DuckDBPyConnection | None = None,
        db: str = ":memory:",
    ) -> None:
        """
        Create a DuckDB database connection and populate the DuckDB database with the
        5 base tables that comprise the Heurist database structure.

        Args:
            hml_xml (bytes): Heurist database structure exported in XML format.
            conn (DuckDBPyConnection | None, optional): A DuckDB database connection. \
                Defaults to None.
            db (str, optional): Path to the DuckDB database. Defaults to ":memory:".
        """

        hml_xml = self.trim_xml_bytes(xml=hml_xml)
        if not conn:
            conn = duckdb.connect(db)
        self.conn = conn
        # Load the Heurist database structure XML into a nested Pydantic data model
        self.hml = HMLStructure.from_xml(hml_xml)

        # Create generic tables
        for t in self.BASE_TABLES:
            name = t[0]
            pydantic_model = getattr(getattr(self.hml, t[1]), t[0])
            self.create(name, pydantic_model)

    @classmethod
    def trim_xml_bytes(cls, xml: bytes) -> bytes:
        """
        Remove any extra whitespace from a potentially malformatted XML.

        Args:
            xml (bytes): Heurist database structure exported XML format.

        Returns:
            bytes: Validated Heurist database structure in XML format.
        """

        return xml.decode("utf-8").strip().encode("utf-8")

    def delete_existing_table(self, table_name: str) -> None:
        """
        If the table already exists in the DuckDB database, drop it.

        Args:
            table_name (str): Name of the table to potentially drop.
        """

        if self.conn.sql(
            f"""
SELECT *
FROM duckdb_tables()
WHERE table_name like '{table_name}'
"""
        ):
            self.conn.sql("DROP TABLE {}".format(table_name))

    def create(self, name: str, model: BaseXmlModel) -> None:
        """Create an empty table in the DuckDB database connection
        based on a Pydantic model.

        Examples:
            >>> # Set up the database class and parse a table model.
            >>> from mock_data import DB_STRUCTURE_XML
            >>> db = HeuristDatabase(hml_xml=DB_STRUCTURE_XML)
            >>> model = db.hml.RecTypeGroups.rtg
            >>>
            >>> # Create a table for the Record Type Group (rtg) table model.
            >>> db.create(name="rtg", model=model)
            >>> shape = db.conn.table("rtg").fetchall()
            >>> # The Record Type Group (rtg) table should have 11 columns.
            >>> len(shape)
            11

        Args:
            model (BaseXmlModel): A Pydantic XML model.
        """

        self.delete_existing_table(name)

        # Convert the model to a dataframe and register it for duckdb
        df = pl.DataFrame(model, infer_schema_length=None)
        assert df.shape[0] > 1

        # Create table from model dataframe
        sql = "CREATE TABLE {} AS FROM df".format(name)
        self.conn.sql(sql)

    def describe_record_schema(self, rty_ID: int) -> DuckDBPyRelation:
        """Join the tables 'dty' (detail), 'rst' (record structure), 'rty' (record type)
        to get all the relevant information for a specific record type, plus add the
        label and description of the section / separator associated with each detail
        (if any).

        Args:
            rty_ID (int): ID of the targeted record type.

        Returns:
            DuckDBPyRelation: A DuckDB Python relation that can be queried or converted.
        """

        return self.conn.from_query(query=RECORD_TYPE_SCHEMA, params=[rty_ID])



from pathlib import Path
import json
from datetime import date
import duckdb

from heurist.api.connection import HeuristAPIConnection
from heurist.database import TransformedDatabase
from heurist.schema.rel_to_dict import convert_rty_description
from heurist.sql.sql_safety import SafeSQLName
from rich.progress import (
    BarColumn,
    MofNCompleteColumn,
    Progress,
    SpinnerColumn,
    TextColumn,
    TimeElapsedColumn,
)

def output_csv(dp: Path, descriptions: list[duckdb.DuckDBPyRelation]) -> None:
    for rel in descriptions:
        name = rel.select("rty_Name").limit(1).fetchone()[0]
        safe_name = SafeSQLName().create_table_name(name)
        fp = dp.joinpath(safe_name).with_suffix(".csv")
        rel.write_csv(file_name=str(fp), header=True)


def output_json(descriptions: list[duckdb.DuckDBPyRelation], fp: Path) -> None:
    date_string = date.today().isoformat()
    data = {"lastModified": date_string, "items": []}
    for desc in descriptions:
        kv_dict = convert_rty_description(description=desc)
        for id, metadata in kv_dict.items():
            d = {"id": id} | metadata
            data["items"].append(d)
    with open(fp, "w", encoding="utf-8") as f:
        json.dump(data, f, indent=4, ensure_ascii=False)


def get_database_schema(
    record_groups: tuple,
    db_name:str,
    login: str,
    password: str,
    debugging: bool,
) -> TransformedDatabase:
    # If testing, load the mock database XML schema
    if debugging:
        from mock_data import DB_STRUCTURE_XML

        db = TransformedDatabase(
            hml_xml=DB_STRUCTURE_XML, record_type_groups=record_groups
        )
    # If not testing, request the database XML schema from the server
    else:
        with (
            Progress(
                TextColumn("{task.description}"),
                SpinnerColumn(),
                TimeElapsedColumn(),
            ) as p,
            HeuristAPIConnection(
                db=db_name,
                login=login,
                password=password,
            ) as client,
        ):
            _ = p.add_task("Downloading schemas")
            xml = client.get_structure()
            db = TransformedDatabase(
                hml_xml=xml,
                record_type_groups=record_groups,
            )
    return db


def export_schema(
    db_name:str,
    login: str,
    password: str,
    record_group: tuple,
    outdir: str,
    output_type: str,
    debugging: bool = False,
):
    # Set up the output directory
    if not outdir:
        outdir = f"{db_name}_schema"
    DIR = Path(outdir)
    DIR.mkdir(exist_ok=True)

    # Get the database schema
    db = get_database_schema(
        record_groups=record_group,
        db_name=db_name,
        login=login,
        password=password,
        debugging=debugging,
    )

    # Describe each targeted record type
    record_type_ids = list(db.pydantic_models.keys())
    with Progress(
        TextColumn("{task.description}"), BarColumn(), MofNCompleteColumn()
    ) as p:
        descriptions = []
        t = p.add_task("Describing record types", total=len(record_type_ids))
        for id in record_type_ids:
            rel = db.describe_record_schema(rty_ID=id)
            descriptions.append(rel)
            p.advance(t)

    # Output the descriptions according to the desired data format
    if output_type == "csv":
        output_csv(dp=DIR, descriptions=descriptions)

    elif output_type == "json":
        outfile = DIR.joinpath("recordTypes.json")
        output_json(descriptions=descriptions, fp=outfile)
